# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Exceptions
===========
"""


from collections import namedtuple
from datetime import timedelta
from enum import Enum
from typing import Any, Optional

import grpc
from google.protobuf.duration_pb2 import Duration

from buildgrid._protos.google.rpc import error_details_pb2, status_pb2


# grpc.Status is a metaclass class, so we derive
# a local _Status and associate the expected Attributes
# with it
class _Status(namedtuple("_Status", ("code", "details", "trailing_metadata")), grpc.Status):
    pass


class ErrorDomain(Enum):
    SERVER = 1
    BOT = 2


class BgdError(Exception):
    """
    Base BuildGrid Error class for internal exceptions.
    """

    def __init__(
        self,
        message: Optional[str],
        *,
        detail: Optional[Any] = None,
        domain: Optional[ErrorDomain] = None,
        reason: Optional[Any] = None,
    ) -> None:
        super().__init__(message)

        # Additional detail and extra information
        self.detail = detail

        # Domand and reason
        self.domain = domain
        self.reason = reason


class ServerError(BgdError):
    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)


class BotError(BgdError):
    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.BOT, reason=reason)


class CancelledError(BgdError):
    """The job was cancelled and any callers should be notified"""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)
        self.last_response = None


class InvalidArgumentError(BgdError):
    """A bad argument was passed, such as a name which doesn't exist."""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)


class NotFoundError(BgdError):
    """Requested resource not found."""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)


class UpdateNotAllowedError(BgdError):
    """UpdateNotAllowedError error."""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)


class OutOfRangeError(BgdError):
    """ByteStream service read data out of range."""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)


class IncompleteReadError(BgdError):
    """ByteStream service read didn't return a full payload."""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)


class FailedPreconditionError(BgdError):
    """One or more errors occurred in setting up the action requested, such as
    a missing input or command or no worker being available. The client may be
    able to fix the errors and retry."""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)


class PermissionDeniedError(BgdError):
    """The caller does not have permission to execute the specified operation."""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)


class BotSessionError(BgdError):
    """Parent class of BotSession Exceptions"""


class BotSessionClosedError(BotSessionError):
    """The requested BotSession has been closed recently."""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)


class UnknownBotSessionError(BotSessionError):
    """Buildgrid does not know the requested BotSession."""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)


class BotSessionMismatchError(BotSessionError):
    """The BotSession details don't match those in BuildGrid's records."""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)


class DuplicateBotSessionError(BotSessionError):
    """The bot with this ID already has a BotSession."""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)


class BotSessionCancelledError(BotSessionError):
    """The BotSession update was cancelled"""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)


class DatabaseError(BgdError):
    """BuildGrid encountered a database error"""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)


class RetriableError(BgdError):
    """BuildGrid encountered a retriable error
    `retry_info` to instruct clients when to retry
    `error_status` a grpc.Status message suitable to call with context.abort_with_status()
    """

    def __init__(
        self, message: str, retry_period: timedelta, detail: Optional[Any] = None, reason: Optional[Any] = None
    ) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)
        retry_delay = Duration()
        retry_delay.FromTimedelta(retry_period)
        retry_info = error_details_pb2.RetryInfo(retry_delay=retry_delay)

        # We could get the integer value of the UNAVAILABLE
        # status code using grpc.StatusCode.UNAVAILABLE.value[0],
        # but the grpc-stubs library complains if we do that. So
        # instead we just hardcode the value, which is unlikely to
        # change as it's a standard code: https://github.com/grpc/grpc/blob/master/doc/statuscodes.md
        status_proto = status_pb2.Status(code=14, message=message)
        error_detail = status_proto.details.add()
        error_detail.Pack(retry_info)

        error_status = _Status(
            code=grpc.StatusCode.UNAVAILABLE,
            details=status_proto.message,
            trailing_metadata=(("grpc-status-details-bin", status_proto.SerializeToString()),),
        )
        self.retry_info = retry_info
        self.error_status = error_status


class RetriableDatabaseError(RetriableError):
    """BuildGrid encountered a retriable database error"""

    def __init__(
        self, message: str, retry_period: timedelta, detail: Optional[Any] = None, reason: Optional[Any] = None
    ) -> None:
        super().__init__(message, retry_period, detail=detail, reason=reason)


class ResourceExhaustedError(BgdError):
    """Some resource has been exhausted, perhaps a per-user quota, or perhaps the entire file system is out of space."""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)


class StorageFullError(ResourceExhaustedError):
    """BuildGrid's storage is full, cannot commit to it."""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, reason=reason)


class GrpcUninitializedError(BgdError):
    """BuildGrid tried to use a gRPC stub before gRPC was initialized."""

    def __init__(self, message: Optional[str], detail: Optional[Any] = None, reason: Optional[Any] = None) -> None:
        super().__init__(message, detail=detail, domain=ErrorDomain.SERVER, reason=reason)



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add worker partial execution timestamps

Revision ID: 38b36022308b
Revises: b5563d55f4e3
Create Date: 2024-10-25 17:04:49.469210

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "38b36022308b"
down_revision = "b5563d55f4e3"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("jobs") as batch_op:
        batch_op.add_column(sa.Column("input_fetch_start_timestamp", sa.DateTime(), nullable=True))
        batch_op.add_column(sa.Column("input_fetch_completed_timestamp", sa.DateTime(), nullable=True))
        batch_op.add_column(sa.Column("output_upload_start_timestamp", sa.DateTime(), nullable=True))
        batch_op.add_column(sa.Column("output_upload_completed_timestamp", sa.DateTime(), nullable=True))
        batch_op.add_column(sa.Column("execution_start_timestamp", sa.DateTime(), nullable=True))
        batch_op.add_column(sa.Column("execution_completed_timestamp", sa.DateTime(), nullable=True))


def downgrade() -> None:
    with op.batch_alter_table("jobs") as batch_op:
        batch_op.drop_column("execution_completed_timestamp")
        batch_op.drop_column("execution_start_timestamp")
        batch_op.drop_column("output_upload_completed_timestamp")
        batch_op.drop_column("output_upload_start_timestamp")
        batch_op.drop_column("input_fetch_completed_timestamp")
        batch_op.drop_column("input_fetch_start_timestamp")



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add command column to the jobs table

Revision ID: 9b595964dc25
Revises: fcb6e8f09a1d
Create Date: 2023-10-05 14:54:18.791620

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9b595964dc25"
down_revision = "e1448dca2c7a"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column("jobs", sa.Column("command", sa.String(), nullable=True))


def downgrade() -> None:
    op.drop_column("jobs", "command")



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Use LargeBinary to store Action message

Revision ID: 8d910c8de8b6
Revises: 2b49634f4459
Create Date: 2020-11-26 16:40:52.682830

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "8d910c8de8b6"
down_revision = "2b49634f4459"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("jobs") as batch_op:
        # Drop this column entirely rather than trying to migrate the content,
        # since anything there is likely to not be usefully recoverable.
        batch_op.drop_column("action")
        # `sa.LargeBinary` selects the actual column type based on the database
        # backend's preference for storing binary data.
        batch_op.add_column(sa.Column("action", sa.LargeBinary(), nullable=True))


def downgrade() -> None:
    with op.batch_alter_table("jobs") as batch_op:
        batch_op.drop_column("action")
        batch_op.add_column(sa.Column("action", sa.String(), nullable=True))



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2020 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import google.protobuf.any_pb2
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.wrappers_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class BuildStatus(google.protobuf.message.Message):
    """Status used for both invocation attempt and overall build completion."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Result:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _ResultEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[BuildStatus._Result.ValueType], builtins.type):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNKNOWN_STATUS: BuildStatus._Result.ValueType  # 0
        """Unspecified or unknown."""
        COMMAND_SUCCEEDED: BuildStatus._Result.ValueType  # 1
        """Build was successful and tests (if requested) all pass."""
        COMMAND_FAILED: BuildStatus._Result.ValueType  # 2
        """Build error and/or test failure."""
        USER_ERROR: BuildStatus._Result.ValueType  # 3
        """Unable to obtain a result due to input provided by the user."""
        SYSTEM_ERROR: BuildStatus._Result.ValueType  # 4
        """Unable to obtain a result due to a failure within the build system."""
        RESOURCE_EXHAUSTED: BuildStatus._Result.ValueType  # 5
        """Build required too many resources, such as build tool RAM."""
        INVOCATION_DEADLINE_EXCEEDED: BuildStatus._Result.ValueType  # 6
        """An invocation attempt time exceeded its deadline."""
        REQUEST_DEADLINE_EXCEEDED: BuildStatus._Result.ValueType  # 8
        """Build request time exceeded the request_deadline"""
        CANCELLED: BuildStatus._Result.ValueType  # 7
        """The build was cancelled by a call to CancelBuild."""

    class Result(_Result, metaclass=_ResultEnumTypeWrapper):
        """The end result of the Build."""

    UNKNOWN_STATUS: BuildStatus.Result.ValueType  # 0
    """Unspecified or unknown."""
    COMMAND_SUCCEEDED: BuildStatus.Result.ValueType  # 1
    """Build was successful and tests (if requested) all pass."""
    COMMAND_FAILED: BuildStatus.Result.ValueType  # 2
    """Build error and/or test failure."""
    USER_ERROR: BuildStatus.Result.ValueType  # 3
    """Unable to obtain a result due to input provided by the user."""
    SYSTEM_ERROR: BuildStatus.Result.ValueType  # 4
    """Unable to obtain a result due to a failure within the build system."""
    RESOURCE_EXHAUSTED: BuildStatus.Result.ValueType  # 5
    """Build required too many resources, such as build tool RAM."""
    INVOCATION_DEADLINE_EXCEEDED: BuildStatus.Result.ValueType  # 6
    """An invocation attempt time exceeded its deadline."""
    REQUEST_DEADLINE_EXCEEDED: BuildStatus.Result.ValueType  # 8
    """Build request time exceeded the request_deadline"""
    CANCELLED: BuildStatus.Result.ValueType  # 7
    """The build was cancelled by a call to CancelBuild."""

    RESULT_FIELD_NUMBER: builtins.int
    FINAL_INVOCATION_ID_FIELD_NUMBER: builtins.int
    BUILD_TOOL_EXIT_CODE_FIELD_NUMBER: builtins.int
    DETAILS_FIELD_NUMBER: builtins.int
    result: global___BuildStatus.Result.ValueType
    """The end result."""
    final_invocation_id: builtins.str
    """Final invocation ID of the build, if there was one.
    This field is only set on a status in BuildFinished event.
    """
    @property
    def build_tool_exit_code(self) -> google.protobuf.wrappers_pb2.Int32Value:
        """Build tool exit code. Integer value returned by the executed build tool.
        Might not be available in some cases, e.g., a build timeout.
        """
    @property
    def details(self) -> google.protobuf.any_pb2.Any:
        """Fine-grained diagnostic information to complement the status."""
    def __init__(
        self,
        *,
        result: global___BuildStatus.Result.ValueType = ...,
        final_invocation_id: builtins.str = ...,
        build_tool_exit_code: google.protobuf.wrappers_pb2.Int32Value | None = ...,
        details: google.protobuf.any_pb2.Any | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["build_tool_exit_code", b"build_tool_exit_code", "details", b"details"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["build_tool_exit_code", b"build_tool_exit_code", "details", b"details", "final_invocation_id", b"final_invocation_id", "result", b"result"]) -> None: ...

global___BuildStatus = BuildStatus



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from contextlib import ExitStack
from typing import Tuple

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import ActionResult, Digest
from buildgrid.server.actioncache.caches.action_cache_abc import ActionCacheABC
from buildgrid.server.exceptions import NotFoundError
from buildgrid.server.logging import buildgrid_logger

LOGGER = buildgrid_logger(__name__)


class WithCacheActionCache(ActionCacheABC):
    """
    An ActionCache that first makes use of a storage and fallback cache

    """

    def __init__(
        self, cache: ActionCacheABC, fallback: ActionCacheABC, allow_updates: bool, cache_failed_actions: bool
    ) -> None:
        """Initialise a new Action Cache with a fallback cache.

        Args:
            cache (ActionCacheABC): Local cache backend instance to be used.
            fallback(ActionCacheABC): Fallback backend instance to be used
            allow_updates(bool): Allow updates pushed to the Action Cache.
                Defaults to ``True``.
            cache_failed_actions(bool): Whether to store failed (non-zero exit
                code) actions. Default to ``True``.
        """
        super().__init__(allow_updates=allow_updates)
        self._stack = ExitStack()
        self._cache_failed_actions = cache_failed_actions
        self._cache = cache
        self._fallback = fallback

    def start(self) -> None:
        self._stack.enter_context(self._cache)
        self._stack.enter_context(self._fallback)

    def stop(self) -> None:
        self._stack.close()

    def get_action_result(self, action_digest: Digest) -> ActionResult:
        """Retrieves the cached result for an Action.

        Will first attempt to retrieve result from cache and then fallback. A
        NotFoundError is raised if both cache and fallback return None.

        Args:
            action_digest (Digest): The digest of the Action to retrieve the
                cached result of.

        """
        cache_result = None
        fallback_result = None
        key = self._get_key(action_digest)
        try:
            cache_result = self._cache.get_action_result(action_digest)
        except NotFoundError:
            pass
        except Exception:
            LOGGER.exception("Unexpected error in cache get_action_result.", tags=dict(digest=action_digest))

        if cache_result is None:
            fallback_result = self._fallback.get_action_result(action_digest)
        else:
            return cache_result

        if fallback_result is not None:
            self._cache.update_action_result(action_digest, fallback_result)
            return fallback_result

        raise NotFoundError(f"Key not found: {key}")

    def update_action_result(self, action_digest: Digest, action_result: ActionResult) -> None:
        """Stores a result for an Action in the cache.

        Will attempt to store result in cache, and then in fallback cache. If
        the result has a non-zero exit code and `cache_failed_actions` is False
        for this cache, the result is not cached.

        Args:
            action_digest (Digest): The digest of the Action whose result is
                being cached.
            action_result (ActionResult): The result to cache for the given
                Action digest.

        """
        if self._cache_failed_actions or action_result.exit_code == 0:
            if not self._allow_updates:
                raise NotImplementedError("Updating cache not allowed")

            try:
                self._cache.update_action_result(action_digest, action_result)

            except Exception:
                LOGGER.warning("Failed to cache action.", tags=dict(digest=action_digest), exc_info=True)

            self._fallback.update_action_result(action_digest, action_result)

    def _get_key(self, action_digest: Digest) -> Tuple[str, int]:
        """Get a hashable cache key from a given Action digest.

        Args:
            action_digest (Digest): The digest to produce a cache key for.

        """
        return (action_digest.hash, action_digest.size_bytes)



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import threading
import time
from contextlib import ExitStack
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest
from buildgrid.server.cas.storage.index.index_abc import IndexABC
from buildgrid.server.context import instance_context
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.metrics_utils import publish_counter_metric, publish_gauge_metric, timer
from buildgrid.server.monitoring import get_monitoring_bus
from buildgrid.server.threading import ContextThreadPoolExecutor, ContextWorker

LOGGER = buildgrid_logger(__name__)


def _digests_str(digests: List[Digest]) -> str:
    return f"{len(digests)} digests ({sum(d.size_bytes for d in digests)} bytes)"


class CASCleanUp:
    """Creates a CAS cleanup service."""

    def __init__(
        self,
        dry_run: bool,
        high_watermark: int,
        low_watermark: int,
        sleep_interval: int,
        batch_size: int,
        only_if_unused_for: timedelta,
        indexes: Dict[str, IndexABC],
        monitor: bool,
    ) -> None:
        self._stack = ExitStack()
        self._dry_run = dry_run

        self._high_watermark = high_watermark
        self._low_watermark = low_watermark
        self._batch_size = batch_size
        self._only_if_unused_for = only_if_unused_for

        self._indexes = indexes

        self._is_instrumented = monitor

        self._sleep_interval = sleep_interval

    # --- Public API ---

    def start(self, timeout: Optional[float] = None) -> None:
        """Start cleanup service"""
        if self._is_instrumented:
            self._stack.enter_context(get_monitoring_bus())
        worker = self._stack.enter_context(ContextWorker(self._begin_cleanup, "CleanUpLauncher"))
        worker.wait(timeout=timeout)

    def stop(self, *args: Any, **kwargs: Any) -> None:
        """Stops the cleanup service"""
        LOGGER.info("Stopping Cleanup Service.")
        self._stack.close()

    def _begin_cleanup(self, stop_requested: threading.Event) -> None:
        if self._dry_run:
            for instance_name in self._indexes.keys():
                self._calculate_cleanup(instance_name)
            return

        attempts = 0
        with ContextThreadPoolExecutor(max_workers=len(self._indexes)) as ex:
            while True:
                futures = {
                    instance_name: ex.submit(self._cleanup_worker, instance_name, stop_requested)
                    for instance_name in self._indexes.keys()
                }

                failed = False
                for instance_name, future in futures.items():
                    try:
                        future.result()
                        LOGGER.info("Cleanup completed.", tags=dict(instance_name=instance_name))
                    except Exception:
                        LOGGER.exception("Cleanup failed.", tags=dict(instance_name=instance_name))
                        failed = True

                if not failed:
                    break

                # Exponential backoff before retrying
                sleep_time = 1.6**attempts
                LOGGER.info("Retrying Cleanup after delay...", tags=dict(sleep_time_seconds=sleep_time))
                stop_requested.wait(timeout=sleep_time)
                attempts += 1
                continue

    def _calculate_cleanup(self, instance_name: str) -> None:
        """Work out which blobs will be deleted by the cleanup command."""
        with instance_context(instance_name):
            LOGGER.info("Cleanup dry run.", tags=dict(instance_name=instance_name))
            index = self._indexes[instance_name]
            only_delete_before = self._get_last_accessed_threshold()
            total_size = index.get_total_size()
            LOGGER.info(
                "Calculated CAS size.",
                tags=dict(
                    total_size=total_size,
                    high_watermark_bytes=self._high_watermark,
                    low_watermark_bytes=self._low_watermark,
                ),
            )
            if total_size >= self._high_watermark:
                required_space = total_size - self._low_watermark
                cleared_space = index.delete_n_bytes(
                    required_space, dry_run=True, protect_blobs_after=only_delete_before
                )
                LOGGER.info(f"Determined {cleared_space} of the requested {required_space} bytes would be deleted.")
            else:
                LOGGER.info(f"Total size {total_size} is less than the high water mark, " f"nothing will be deleted.")

    def _do_cleanup_batch(
        self,
        instance_name: str,
        index: IndexABC,
        only_delete_before: datetime,
        total_size: int,
        stop_requested: threading.Event,
    ) -> None:
        batch_start_time = time.time()

        LOGGER.info("Deleting bytes from the index.", tags=dict(batch_size=self._batch_size))
        bytes_deleted = index.delete_n_bytes(self._batch_size, protect_blobs_after=only_delete_before)

        if not bytes_deleted:
            err = (
                "Marked 0 digests for deletion, even though cleanup was triggered. "
                "This may be because the remaining digests have been accessed within "
                f"{only_delete_before}."
            )
            if total_size >= self._high_watermark:
                LOGGER.error(f"{err} Total size still remains greater than high watermark!")
            else:
                LOGGER.warning(err)
            stop_requested.wait(timeout=self._sleep_interval)  # Avoid a busy loop when we can't make progress
            return

        LOGGER.info("Bulk deleted bytes from index.", tags=dict(bytes_deleted=bytes_deleted))

        if self._is_instrumented:
            batch_duration = time.time() - batch_start_time
            bytes_deleted_per_second = bytes_deleted / batch_duration
            publish_gauge_metric(METRIC.CLEANUP.BYTES_DELETED_PER_SECOND, bytes_deleted_per_second)
            publish_counter_metric(METRIC.CLEANUP.BYTES_DELETED_COUNT, bytes_deleted)

    def _cleanup_worker(self, instance_name: str, stop_requested: threading.Event) -> None:
        """Cleanup when full"""
        with instance_context(instance_name):
            index = self._indexes[instance_name]
            LOGGER.info("Cleanup started.", tags=dict(instance_name=instance_name))

            while not stop_requested.is_set():
                # When first starting a loop, we will also include any remaining delete markers as part of
                # the total size.
                total_size = index.get_total_size()
                self.publish_total_size_metric(total_size)

                if total_size >= self._high_watermark:
                    to_delete = total_size - self._low_watermark
                    LOGGER.info(
                        "High watermark exceeded. Deleting items from storage/index.",
                        tags=dict(total_bytes=total_size, min_bytes_to_delete=to_delete, instance_name=instance_name),
                    )

                    with timer(METRIC.CLEANUP.DURATION):
                        while not stop_requested.is_set() and total_size > self._low_watermark:
                            only_delete_before = self._get_last_accessed_threshold()
                            with timer(METRIC.CLEANUP.BATCH_DURATION):
                                self._do_cleanup_batch(
                                    instance_name=instance_name,
                                    index=index,
                                    only_delete_before=only_delete_before,
                                    total_size=total_size,
                                    stop_requested=stop_requested,
                                )
                            total_size = index.get_total_size()
                            self.publish_total_size_metric(total_size)
                            LOGGER.info("Finished cleanup batch.", tags=dict(non_stale_total_bytes=total_size))

                    LOGGER.info("Finished cleanup.", tags=dict(total_bytes=total_size))

                stop_requested.wait(timeout=self._sleep_interval)

    def _get_last_accessed_threshold(self) -> datetime:
        return datetime.utcnow() - self._only_if_unused_for

    def publish_total_size_metric(self, total_size: int) -> None:
        if self._is_instrumented:
            publish_gauge_metric(METRIC.CLEANUP.TOTAL_BYTES_COUNT, total_size)
            publish_gauge_metric(METRIC.CLEANUP.LOW_WATERMARK_BYTES_COUNT, self._low_watermark)
            publish_gauge_metric(METRIC.CLEANUP.HIGH_WATERMARK_BYTES_COUNT, self._high_watermark)



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Operations command
=================

Check the status of operations
"""

import sys
from textwrap import indent
from typing import Any

import click
from google.protobuf import json_format
from grpc import RpcError, StatusCode

from buildgrid._protos.build.bazel.remote.execution.v2 import remote_execution_pb2, remote_execution_pb2_grpc
from buildgrid._protos.google.longrunning import operations_pb2, operations_pb2_grpc
from buildgrid._protos.google.longrunning.operations_pb2 import Operation
from buildgrid._protos.google.rpc import code_pb2
from buildgrid.server.client.channel import setup_channel
from buildgrid.server.enums import OperationStage
from buildgrid.server.exceptions import InvalidArgumentError
from buildgrid.server.metadata import extract_request_metadata, printable_request_metadata

from ..cli import pass_context
from .rpc_utils import cancel_operation


@click.group(name="operation", short_help="Long running operations commands.")
@click.option(
    "--remote",
    type=click.STRING,
    default="http://localhost:50051",
    show_default=True,
    help="Remote execution server's URL (port defaults to 50051 if no specified).",
)
@click.option(
    "--auth-token",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Authorization token for the remote.",
)
@click.option(
    "--client-key",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Private client key for TLS (PEM-encoded).",
)
@click.option(
    "--client-cert",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Public client certificate for TLS (PEM-encoded).",
)
@click.option(
    "--server-cert",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Public server certificate for TLS (PEM-encoded).",
)
@click.option(
    "--instance-name", type=click.STRING, default=None, show_default=True, help="Targeted farm instance name."
)
@click.option("--action-id", type=str, help="Action ID.")
@click.option("--invocation-id", type=str, help="Tool invocation ID.")
@click.option("--correlation-id", type=str, help="Correlated invocation ID.")
@pass_context
def cli(
    context: Any,
    remote: str,
    instance_name: str,
    auth_token: str,
    client_key: str,
    client_cert: str,
    server_cert: str,
    action_id: str,
    invocation_id: str,
    correlation_id: str,
) -> None:
    """Entry point for the bgd-operation CLI command group."""
    try:
        context.channel, _ = setup_channel(
            remote,
            auth_token=auth_token,
            client_key=client_key,
            client_cert=client_cert,
            server_cert=server_cert,
            action_id=action_id,
            tool_invocation_id=invocation_id,
            correlated_invocations_id=correlation_id,
        )

    except InvalidArgumentError as e:
        click.echo(f"Error: {e}.", err=True)
        sys.exit(-1)

    context.instance_name = instance_name


def _print_operation_status(operation: Operation, print_details: bool = False) -> None:
    op_metadata = remote_execution_pb2.ExecuteOperationMetadata()
    # The metadata is expected to be an ExecuteOperationMetadata message:
    if not operation.metadata.Is(op_metadata.DESCRIPTOR):
        raise InvalidArgumentError("Metadata is not an ExecuteOperationMetadata message")
    operation.metadata.Unpack(op_metadata)

    stage = OperationStage(op_metadata.stage)

    if not operation.done:
        if stage == OperationStage.CACHE_CHECK:
            click.echo(f"CacheCheck: {operation.name}: " f"Querying action-cache (stage={op_metadata.stage})")
        elif stage == OperationStage.QUEUED:
            click.echo(f"Queued: {operation.name}: " f"Waiting for execution (stage={op_metadata.stage})")
        elif stage == OperationStage.EXECUTING:
            click.echo(f"Executing: {operation.name}: " f"Currently running (stage={op_metadata.stage})")
        else:
            click.echo(f"Error: {operation.name}: " f"In an invalid state (stage={op_metadata.stage})", err=True)
        return

    assert stage == OperationStage.COMPLETED

    response = remote_execution_pb2.ExecuteResponse()
    # The response is expected to be an ExecutionResponse message:
    operation.response.Unpack(response)

    if response.status.code != code_pb2.OK:
        click.echo(f"Failure: {operation.name}: {response.status.message} " f"(code={response.status.code})")
    else:
        if response.result.exit_code != 0:
            click.echo(
                f"Success: {operation.name}: Completed with failure "
                f"(stage={op_metadata.stage}, exit_code={response.result.exit_code})"
            )
        else:
            click.echo(
                f"Success: {operation.name}: Completed succesfully "
                f"(stage={op_metadata.stage}, exit_code={response.result.exit_code})"
            )

    if print_details:
        act_metadata = response.result.execution_metadata
        click.echo(indent(f"worker={act_metadata.worker}", "  "))

        queued = act_metadata.queued_timestamp.ToDatetime()
        click.echo(indent(f"queued_at={queued}", "  "))

        worker_start = act_metadata.worker_start_timestamp.ToDatetime()
        worker_completed = act_metadata.worker_completed_timestamp.ToDatetime()
        click.echo(indent(f"work_duration={worker_completed - worker_start}", "  "))

        fetch_start = act_metadata.input_fetch_start_timestamp.ToDatetime()
        fetch_completed = act_metadata.input_fetch_completed_timestamp.ToDatetime()
        click.echo(indent(f"fetch_duration={fetch_completed - fetch_start}", "    "))

        execution_start = act_metadata.execution_start_timestamp.ToDatetime()
        execution_completed = act_metadata.execution_completed_timestamp.ToDatetime()
        click.echo(indent(f"exection_duration={execution_completed - execution_start}", "    "))

        upload_start = act_metadata.output_upload_start_timestamp.ToDatetime()
        upload_completed = act_metadata.output_upload_completed_timestamp.ToDatetime()
        click.echo(indent(f"upload_duration={upload_completed - upload_start}", "    "))

        click.echo(indent(f"total_duration={worker_completed - queued}", "  "))


@cli.command("status", short_help="Get the status of an operation.")
@click.argument("operation-name", nargs=1, type=click.STRING, required=True)
@click.option("--json", is_flag=True, show_default=True, help="Print operations status in JSON format.")
@click.option("--show-request-metadata", is_flag=True, show_default=False, help="Show RequestMetadata message")
@pass_context
def status(context: Any, operation_name: str, json: bool, show_request_metadata: bool) -> None:
    stub = operations_pb2_grpc.OperationsStub(context.channel)
    request = operations_pb2.GetOperationRequest(name=operation_name)

    try:
        operation, call = stub.GetOperation.with_call(request)
    except RpcError as e:
        click.echo(f"Error: {e.details()}", err=True)
        sys.exit(-1)

    if not json:
        _print_operation_status(operation, print_details=True)
    else:
        click.echo(json_format.MessageToJson(operation))

    if show_request_metadata:
        metadata = call.trailing_metadata()
        if json:
            request_metadata_proto = extract_request_metadata(metadata)
            click.echo(json_format.MessageToJson(request_metadata_proto))
        else:
            click.echo(f"Request metadata: [{printable_request_metadata(call.trailing_metadata())}]")


@cli.command("cancel", short_help="Cancel an operation.")
@click.argument("operation-name", nargs=1, type=click.STRING, required=True)
@pass_context
def cancel(context: Any, operation_name: str) -> None:
    cancel_successful = cancel_operation(context, operation_name)
    if not cancel_successful:
        sys.exit(-1)


@cli.command("list", short_help="List operations.")
@click.option("--json", is_flag=True, show_default=True, help="Print operations list in JSON format.")
@click.option(
    "--page-token",
    type=click.STRING,
    help='Token to start from. Set this to the "next_page_token" of the previous '
    "ListOperationsResult to get the next page of results.",
)
@click.option(
    "--page-size",
    type=click.INT,
    help="Number of operations to request per page. Cannot be larger than the server maximum (throws an error).",
)
@click.option(
    "--filter-string",
    "--filter",
    type=click.STRING,
    default="",
    help="Filter which operations are returned. By default, only incomplete operations are shown. "
    "See docs for syntax.",
)
@pass_context
def lists(context: Any, json: bool, page_token: str, page_size: int, filter_string: str) -> None:
    stub = operations_pb2_grpc.OperationsStub(context.channel)
    request = operations_pb2.ListOperationsRequest(
        name=context.instance_name, page_token=page_token, page_size=page_size, filter=filter_string
    )

    try:
        response = stub.ListOperations(request)
    except RpcError as e:
        click.echo(f"Error: {e.details()}", err=True)
        sys.exit(-1)

    if not response.operations:
        click.echo("Error: No operations to list.", err=True)
        return

    for operation in response.operations:
        if not json:
            _print_operation_status(operation)
        else:
            click.echo(json_format.MessageToJson(operation))

    if response.next_page_token:
        next_cmd = "bgd operation list "
        if filter_string:
            next_cmd = next_cmd + f"--filter {filter_string} "
        if page_size:
            next_cmd = next_cmd + f"--page-size {page_size} "
        next_cmd = next_cmd + f'--page-token "{response.next_page_token}"'
        click.echo(f"Get the next page of operations with [{next_cmd}]")


@cli.command("wait", short_help="Streams an operation until it is complete.")
@click.argument("operation-name", nargs=1, type=click.STRING, required=True)
@click.option("--json", is_flag=True, show_default=True, help="Print operations statuses in JSON format.")
@pass_context
def wait(context: Any, operation_name: str, json: bool) -> None:
    stub = remote_execution_pb2_grpc.ExecutionStub(context.channel)
    request = remote_execution_pb2.WaitExecutionRequest(name=operation_name)

    operation_iterator = stub.WaitExecution(request)

    try:
        for operation in operation_iterator:
            if not json and operation.done:
                _print_operation_status(operation, print_details=True)
            elif not json:
                _print_operation_status(operation)
            else:
                click.echo(json_format.MessageToJson(operation))

    except InvalidArgumentError as e:
        click.echo(f"Error: In the reply: {e}", err=True)
        sys.exit(-1)

    except RpcError as e:
        if e.code() != StatusCode.CANCELLED:
            click.echo(f"Error: {e.details()}", err=True)
            sys.exit(-1)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2016 Google Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import google.protobuf.any_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import buildgrid._protos.google.rpc.status_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class Operation(google.protobuf.message.Message):
    """This resource represents a long-running operation that is the result of a
    network API call.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    DONE_FIELD_NUMBER: builtins.int
    ERROR_FIELD_NUMBER: builtins.int
    RESPONSE_FIELD_NUMBER: builtins.int
    name: builtins.str
    """The server-assigned name, which is only unique within the same service that
    originally returns it. If you use the default HTTP mapping, the
    `name` should have the format of `operations/some/unique/name`.
    """
    @property
    def metadata(self) -> google.protobuf.any_pb2.Any:
        """Service-specific metadata associated with the operation.  It typically
        contains progress information and common metadata such as create time.
        Some services might not provide such metadata.  Any method that returns a
        long-running operation should document the metadata type, if any.
        """
    done: builtins.bool
    """If the value is `false`, it means the operation is still in progress.
    If true, the operation is completed, and either `error` or `response` is
    available.
    """
    @property
    def error(self) -> buildgrid._protos.google.rpc.status_pb2.Status:
        """The error result of the operation in case of failure or cancellation."""
    @property
    def response(self) -> google.protobuf.any_pb2.Any:
        """The normal response of the operation in case of success.  If the original
        method returns no data on success, such as `Delete`, the response is
        `google.protobuf.Empty`.  If the original method is standard
        `Get`/`Create`/`Update`, the response should be the resource.  For other
        methods, the response should have the type `XxxResponse`, where `Xxx`
        is the original method name.  For example, if the original method name
        is `TakeSnapshot()`, the inferred response type is
        `TakeSnapshotResponse`.
        """
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        metadata: google.protobuf.any_pb2.Any | None = ...,
        done: builtins.bool = ...,
        error: buildgrid._protos.google.rpc.status_pb2.Status | None = ...,
        response: google.protobuf.any_pb2.Any | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["error", b"error", "metadata", b"metadata", "response", b"response", "result", b"result"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["done", b"done", "error", b"error", "metadata", b"metadata", "name", b"name", "response", b"response", "result", b"result"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["result", b"result"]) -> typing_extensions.Literal["error", "response"] | None: ...

global___Operation = Operation

@typing_extensions.final
class GetOperationRequest(google.protobuf.message.Message):
    """The request message for [Operations.GetOperation][google.longrunning.Operations.GetOperation]."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    name: builtins.str
    """The name of the operation resource."""
    def __init__(
        self,
        *,
        name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["name", b"name"]) -> None: ...

global___GetOperationRequest = GetOperationRequest

@typing_extensions.final
class ListOperationsRequest(google.protobuf.message.Message):
    """The request message for [Operations.ListOperations][google.longrunning.Operations.ListOperations]."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    FILTER_FIELD_NUMBER: builtins.int
    PAGE_SIZE_FIELD_NUMBER: builtins.int
    PAGE_TOKEN_FIELD_NUMBER: builtins.int
    name: builtins.str
    """The name of the operation collection."""
    filter: builtins.str
    """The standard list filter."""
    page_size: builtins.int
    """The standard list page size."""
    page_token: builtins.str
    """The standard list page token."""
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        filter: builtins.str = ...,
        page_size: builtins.int = ...,
        page_token: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["filter", b"filter", "name", b"name", "page_size", b"page_size", "page_token", b"page_token"]) -> None: ...

global___ListOperationsRequest = ListOperationsRequest

@typing_extensions.final
class ListOperationsResponse(google.protobuf.message.Message):
    """The response message for [Operations.ListOperations][google.longrunning.Operations.ListOperations]."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    OPERATIONS_FIELD_NUMBER: builtins.int
    NEXT_PAGE_TOKEN_FIELD_NUMBER: builtins.int
    @property
    def operations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Operation]:
        """A list of operations that matches the specified filter in the request."""
    next_page_token: builtins.str
    """The standard List next-page token."""
    def __init__(
        self,
        *,
        operations: collections.abc.Iterable[global___Operation] | None = ...,
        next_page_token: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["next_page_token", b"next_page_token", "operations", b"operations"]) -> None: ...

global___ListOperationsResponse = ListOperationsResponse

@typing_extensions.final
class CancelOperationRequest(google.protobuf.message.Message):
    """The request message for [Operations.CancelOperation][google.longrunning.Operations.CancelOperation]."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    name: builtins.str
    """The name of the operation resource to be cancelled."""
    def __init__(
        self,
        *,
        name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["name", b"name"]) -> None: ...

global___CancelOperationRequest = CancelOperationRequest

@typing_extensions.final
class DeleteOperationRequest(google.protobuf.message.Message):
    """The request message for [Operations.DeleteOperation][google.longrunning.Operations.DeleteOperation]."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    name: builtins.str
    """The name of the operation resource to be deleted."""
    def __init__(
        self,
        *,
        name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["name", b"name"]) -> None: ...

global___DeleteOperationRequest = DeleteOperationRequest



# Copyright (C) 2019 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
RedisStorage
==================

A storage provider that stores data in a persistent redis store.
https://redis.io/

Redis client: redis-py
https://github.com/andymccurdy/redis-py

"""
import functools
import io
from typing import IO, Any, Callable, Dict, List, Optional, Tuple, TypeVar, cast

import redis

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest
from buildgrid._protos.google.rpc import code_pb2
from buildgrid._protos.google.rpc.status_pb2 import Status
from buildgrid.server.decorators import timed
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.redis.provider import RedisProvider
from buildgrid.server.settings import HASH

from .storage_abc import StorageABC

LOGGER = buildgrid_logger(__name__)

Func = TypeVar("Func", bound=Callable)  # type: ignore[type-arg]


def redis_client_exception_wrapper(func: Func) -> Func:
    """Wrapper from handling redis client exceptions."""

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except redis.RedisError:
            LOGGER.exception("Redis Exception.", tags=dict(func_name=func.__name__))
            raise RuntimeError

    return cast(Func, wrapper)


class RedisStorage(StorageABC):
    """Interface for communicating with a redis store."""

    TYPE = "Redis"

    @redis_client_exception_wrapper
    def __init__(self, redis: RedisProvider) -> None:
        self._redis = redis

    def _construct_key(self, digest: Digest) -> str:
        """Helper to get the redis key name for a particular digest"""
        return digest.hash + "_" + str(digest.size_bytes)

    @timed(METRIC.STORAGE.STAT_DURATION, type=TYPE)
    @redis_client_exception_wrapper
    def has_blob(self, digest: Digest) -> bool:
        LOGGER.debug("Checking for blob.", tags=dict(digest=digest))
        return bool(self._redis.execute_ro(lambda r: r.exists(self._construct_key(digest))))

    @timed(METRIC.STORAGE.READ_DURATION, type=TYPE)
    @redis_client_exception_wrapper
    def get_blob(self, digest: Digest) -> Optional[IO[bytes]]:
        LOGGER.debug("Getting blob.", tags=dict(digest=digest))
        blob = self._redis.execute_ro(lambda r: r.get(self._construct_key(digest)))
        return None if blob is None else io.BytesIO(blob)

    @timed(METRIC.STORAGE.DELETE_DURATION, type=TYPE)
    @redis_client_exception_wrapper
    def delete_blob(self, digest: Digest) -> None:
        LOGGER.debug("Deleting blob.", tags=dict(digest=digest))
        self._redis.execute_rw(lambda r: r.delete(self._construct_key(digest)))

    @timed(METRIC.STORAGE.DELETE_DURATION, type=TYPE)
    @redis_client_exception_wrapper
    def bulk_delete(self, digests: List[Digest]) -> List[str]:
        keys = [self._construct_key(digest) for digest in digests]
        self._redis.execute_rw(lambda r: r.delete(*keys))
        return []

    @timed(METRIC.STORAGE.WRITE_DURATION, type=TYPE)
    @redis_client_exception_wrapper
    def commit_write(self, digest: Digest, write_session: IO[bytes]) -> None:
        LOGGER.debug("Writing blob.", tags=dict(digest=digest))
        write_session.seek(0)
        self._redis.execute_rw(lambda r: r.set(self._construct_key(digest), write_session.read()))

    @timed(METRIC.STORAGE.BULK_STAT_DURATION, type=TYPE)
    @redis_client_exception_wrapper
    def missing_blobs(self, digests: List[Digest]) -> List[Digest]:
        # Exist takes multiple keys, but only returns the number of keys which
        # exist, not which keys do/don't exist. Instead pipeline N exist
        # calls, which allows distinguishing which keys do/don't exist.

        def validate_digests(r: "redis.Redis[bytes]") -> List[int]:
            pipe = r.pipeline()
            for digest in digests:
                pipe.exists(self._construct_key(digest))
            return pipe.execute()

        results = self._redis.execute_ro(validate_digests)

        missing_digests: List[Digest] = []
        for digest, result in zip(digests, results):
            if not result:
                missing_digests.append(digest)
        return missing_digests

    @timed(METRIC.STORAGE.BULK_WRITE_DURATION, type=TYPE)
    @redis_client_exception_wrapper
    def bulk_update_blobs(self, blobs: List[Tuple[Digest, bytes]]) -> List[Status]:
        keymap: Dict[str, bytes] = {}
        results: List[Status] = []
        for digest, data in blobs:
            if len(data) != digest.size_bytes or HASH(data).hexdigest() != digest.hash:
                results.append(Status(code=code_pb2.INVALID_ARGUMENT, message="Data doesn't match hash"))
                continue
            results.append(Status(code=code_pb2.OK))
            keymap[self._construct_key(digest)] = data

        self._redis.execute_rw(lambda r: r.mset(keymap))  # type: ignore[arg-type]
        # mset can't fail according to the documentation so return OK for all remaining digests
        return results

    @timed(METRIC.STORAGE.BULK_READ_DURATION, type=TYPE)
    @redis_client_exception_wrapper
    def bulk_read_blobs(self, digests: List[Digest]) -> Dict[str, bytes]:
        redis_keys = [self._construct_key(x) for x in digests]
        found_blobs = self._redis.execute_ro(lambda r: r.mget(redis_keys))
        result_map: Dict[str, bytes] = {}
        for digest, blob in zip(digests, found_blobs):
            if blob is not None:
                result_map[digest.hash] = blob
        return result_map



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import ActionResult, Digest
from buildgrid.server.actioncache.caches.action_cache_abc import ActionCacheABC
from buildgrid.server.exceptions import NotFoundError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.metrics_utils import publish_counter_metric

LOGGER = buildgrid_logger(__name__)


class MirroredCache(ActionCacheABC):
    """Synchronize two mirrored action-caches to the same state"""

    def __init__(
        self,
        first: ActionCacheABC,
        second: ActionCacheABC,
    ):
        # Don't pass a storage object to super class
        # as blob existence check will be performed by each object.
        # On the other hand, the storages of these two caches should point to the same one
        # or also mirrored with each other.
        super().__init__(allow_updates=True, storage=None)
        self._first = first
        self._second = second

    def start(self) -> None:
        self._first.start()
        self._second.start()

    def stop(self) -> None:
        self._second.stop()
        self._first.stop()

    def update_action_result(self, action_digest: Digest, action_result: ActionResult) -> None:
        self._first.update_action_result(action_digest, action_result)
        self._second.update_action_result(action_digest, action_result)
        LOGGER.info("Finished dual write to both action-caches.", tags=dict(digest=action_digest))

    def get_action_result(self, action_digest: Digest) -> ActionResult:
        first_result = _try_get_action_result(self._first, action_digest)
        second_result = _try_get_action_result(self._second, action_digest)

        if first_result is None and second_result is None:
            raise NotFoundError(f"Action result not found: {action_digest.hash}/{action_digest.size_bytes}")

        if first_result is None:
            publish_counter_metric(METRIC.ACTION_CACHE.MIRRORED_MISMATCH_COUNT, 1)
            self._first.update_action_result(action_digest, second_result)  # type: ignore[arg-type]
            return second_result  # type: ignore[return-value]

        if second_result is None:
            publish_counter_metric(METRIC.ACTION_CACHE.MIRRORED_MISMATCH_COUNT, 1)
            self._second.update_action_result(action_digest, first_result)
            return first_result

        if first_result != second_result:
            publish_counter_metric(METRIC.ACTION_CACHE.MIRRORED_MISMATCH_COUNT, 1)
            LOGGER.warning(
                "Different action results in mirrored caches.",
                tags=dict(first_result=first_result, second_result=second_result),
            )
            return first_result

        publish_counter_metric(METRIC.ACTION_CACHE.MIRRORED_MATCH_COUNT, 1)
        return first_result


def _try_get_action_result(cache: ActionCacheABC, digest: Digest) -> Optional[ActionResult]:
    try:
        return cache.get_action_result(digest)
    except NotFoundError:
        return None



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import sys
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Hashable,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypedDict,
    TypeVar,
    Union,
)
from urllib.parse import urlparse

import buildgrid_metering.client as metering
import click
import grpc
import jsonschema
import requests
import yaml
from buildgrid_metering.client.exceptions import MeteringServiceClientError, MeteringServiceError
from importlib_resources import files

from buildgrid.server.actioncache.caches.action_cache_abc import ActionCacheABC
from buildgrid.server.actioncache.caches.lru_cache import LruActionCache
from buildgrid.server.actioncache.caches.mirrored_cache import MirroredCache
from buildgrid.server.actioncache.caches.remote_cache import RemoteActionCache
from buildgrid.server.actioncache.caches.s3_cache import S3ActionCache
from buildgrid.server.actioncache.caches.with_cache import WithCacheActionCache
from buildgrid.server.actioncache.caches.write_once_cache import WriteOnceActionCache
from buildgrid.server.actioncache.instance import ActionCache
from buildgrid.server.bots.instance import BotsInterface
from buildgrid.server.build_events.storage import BuildEventStreamStorage
from buildgrid.server.cas.instance import ByteStreamInstance, ContentAddressableStorageInstance
from buildgrid.server.cas.storage.disk import DiskStorage
from buildgrid.server.cas.storage.index.sql import SQLIndex
from buildgrid.server.cas.storage.lru_memory_cache import LRUMemoryCache
from buildgrid.server.cas.storage.remote import RemoteStorage
from buildgrid.server.cas.storage.replicated import ReplicatedStorage
from buildgrid.server.cas.storage.s3 import S3Storage
from buildgrid.server.cas.storage.sharded import ShardedStorage
from buildgrid.server.cas.storage.size_differentiated import SizeDifferentiatedStorage, SizeLimitedStorageType
from buildgrid.server.cas.storage.sql import SQLStorage
from buildgrid.server.cas.storage.storage_abc import StorageABC
from buildgrid.server.cas.storage.with_cache import WithCacheStorage
from buildgrid.server.client.asset import AssetClient
from buildgrid.server.client.authentication import ClientCredentials
from buildgrid.server.client.channel import setup_channel
from buildgrid.server.controller import ExecutionController
from buildgrid.server.enums import ActionCacheEntryType, MeteringThrottleAction, ServiceName
from buildgrid.server.introspection.instance import IntrospectionInstance
from buildgrid.server.scheduler import (
    AgedJobHandlerOptions,
    DynamicPropertySet,
    PropertyLabel,
    PropertySet,
    Scheduler,
    StaticPropertySet,
)
from buildgrid.server.settings import (
    DEFAULT_MAX_EXECUTION_TIMEOUT,
    DEFAULT_MAX_LIST_OPERATION_PAGE_SIZE,
    DEFAULT_PLATFORM_PROPERTY_KEYS,
    INSECURE_URI_SCHEMES,
    S3_MAX_RETRIES,
    S3_TIMEOUT_CONNECT,
    S3_TIMEOUT_READ,
    S3_USERAGENT_NAME,
    SECURE_URI_SCHEMES,
)
from buildgrid.server.sql.provider import SqlProvider

from .config import ChannelConfig

if TYPE_CHECKING:
    from buildgrid.server.actioncache.caches.redis_cache import RedisActionCache
    from buildgrid.server.cas.storage.index.redis import RedisIndex
    from buildgrid.server.cas.storage.redis import RedisStorage
    from buildgrid.server.redis.provider import RedisProvider


_Func = TypeVar("_Func", bound=Callable)  # type: ignore[type-arg]

# Stores the definitions of struct loaders for tags.
object_definitions: Dict[str, Callable] = {}  # type: ignore[type-arg]

# Stores notes for keys that are marked deprecated. key=yaml-key, value=deprecation-message
deprecated_object_keys: Dict[str, Dict[str, str]] = {}

# Stores the definitions of string loaders for tags
string_definitions: Dict[str, Callable] = {}  # type: ignore[type-arg]


def object_tag(kind: str, deprecated_keys: Optional[Dict[str, str]] = None) -> Callable[[_Func], _Func]:
    """
    Register a tag with custom decoder logic for a yaml object field.
    """

    def wrapper(f: _Func) -> _Func:
        object_definitions[kind] = f
        if deprecated_keys is not None:
            deprecated_object_keys[kind] = deprecated_keys
        return f

    return wrapper


def string_tag(kind: str) -> Callable[[_Func], _Func]:
    """
    Register a tag with custom decoder logic for a yaml string value field.
    """

    def wrapper(f: _Func) -> _Func:
        string_definitions[kind] = f
        return f

    return wrapper


@object_tag("!channel")
def load_channel(
    insecure_mode: bool,
    address: str,
    credentials: Optional[Dict[str, str]] = None,
) -> ChannelConfig:
    """Creates a GRPC channel.

    The :class:`Channel` class returns a `grpc.Channel` and is generated from
    the tag ``!channel``. Creates either a secure or insecure channel.

    Usage
        .. code:: yaml

            - !channel
              address (str): Address for the channel. (For example,
                'localhost:50055' or 'unix:///tmp/sock')
              port (int): A port for the channel (only if no address was specified).
              insecure-mode: false
              credentials:
                tls-server-key: !expand-path ~/.config/buildgrid/server.key
                tls-server-cert: !expand-path ~/.config/buildgrid/server.cert
                tls-client-certs: !expand-path ~/.config/buildgrid/client.cert

    Args:
        port (int): A port for the channel.
        insecure_mode (bool): If ``True``, generates an insecure channel, even
            if there are credentials. Defaults to ``True``.
        credentials (dict, optional): A dictionary in the form::

            tls-server-key: /path/to/server-key
            tls-server-cert: /path/to/server-cert
            tls-client-certs: /path/to/client-certs
    """

    if not insecure_mode:
        _validate_server_credentials(credentials)
        return ChannelConfig(insecure_mode=insecure_mode, address=address, credentials=credentials)
    return ChannelConfig(insecure_mode=False, address=address, credentials=None)


@string_tag("!expand-path")
def expand_path(path: str) -> str:
    """Returns a string of the user's path after expansion.

    The :class:`ExpandPath` class returns a string and is generated from the
    tag ``!expand-path``.

    Usage
        .. code:: yaml

            path: !expand-path ~/bgd-data/cas

    Args:
        path (str): Can be used with strings such as: ``~/dir/to/something``
            or ``$HOME/certs``
    """

    path = os.path.expanduser(path)
    path = os.path.expandvars(path)
    return path


@string_tag("!expand-vars")
def expand_vars(value: str) -> str:
    """Expand environment variables in a string.

    The :class:`ExpandVars` class returns a string and is generated from the
    tag ``!expand-vars``.

    Usage
        .. code:: yaml

            endpoint: !expand-vars $ENDPOINT

    Args:
        path (str): Can be used with strings such as: ``http://$ENDPOINT``
    """

    return os.path.expandvars(value)


@string_tag("!read-file")
def read_file(path: str) -> str:
    """Returns a string of the contents of the specified file.

    The :class:`ReadFile` class returns a string and is generated from the
    tag ``!read-file``.

    Usage
        .. code:: yaml

            secret_key: !read-file /var/bgd/s3-secret-key

    Args:
        path (str): Can be used with strings such as: ``~/path/to/some/file``
            or ``$HOME/myfile`` or ``/path/to/file``
    """

    path = os.path.expandvars(os.path.expanduser(path))

    if not os.path.exists(path):
        click.echo(
            click.style(
                f"ERROR: read-file `{path}` failed due to it not existing or bad permissions.",
                fg="red",
                bold=True,
            ),
            err=True,
        )
        sys.exit(-1)
    else:
        with open(path, "r", encoding="utf-8") as file:
            try:
                file_contents = "\n".join(file.readlines()).strip()
                return file_contents
            except IOError as e:
                click.echo(f"ERROR: read-file failed to read file `{path}`: {e}", err=True)
                sys.exit(-1)


@object_tag("!disk-storage")
def load_disk_storage(path: str) -> DiskStorage:
    """Generates :class:`buildgrid.server.cas.storage.disk.DiskStorage` using the tag ``!disk-storage``.

    Usage
        .. code:: yaml

            - !disk-storage
              path: /opt/bgd/cas-storage

    Args:
        path (str): Path to directory to storage.
    """

    return DiskStorage(path)


@object_tag("!lru-storage")
def load_lru_storage(size: str) -> LRUMemoryCache:
    """Generates :class:`buildgrid.server.cas.storage.lru_memory_cache.LRUMemoryCache` using the tag ``!lru-storage``.

    Usage
        .. code:: yaml

            - !lru-storage
              size: 2048M

    Args:
        size (int): Size e.g ``10kb``. Size parsed with
            :meth:`buildgrid.server.app.settings.parser._parse_size`.
    """

    return LRUMemoryCache(_parse_size(size))


@object_tag("!s3-storage")
def load_s3_storage(
    bucket: str,
    endpoint: str,
    access_key: str,
    secret_key: str,
    read_timeout_seconds_per_kilobyte: Optional[float] = None,
    write_timeout_seconds_per_kilobyte: Optional[float] = None,
    read_timeout_min_seconds: float = S3_TIMEOUT_READ,
    write_timeout_min_seconds: float = S3_TIMEOUT_READ,
    versioned_deletes: bool = False,
    hash_prefix_size: Optional[int] = None,
    path_prefix_string: Optional[str] = None,
) -> S3Storage:
    """Generates :class:`buildgrid.server.cas.storage.s3.S3Storage` using the tag ``!s3-storage``.

    Usage
        .. code:: yaml

            - !s3-storage
              bucket: bgd-bucket-{digest[0]}{digest[1]}
              endpoint: http://127.0.0.1:9000
              access_key: !read-file /var/bgd/s3-access-key
              secret_key: !read-file /var/bgd/s3-secret-key
              read_timeout_seconds_per_kilobyte: 0.01
              write_timeout_seconds_per_kilobyte: 0.01
              read_timeout_min_seconds: 120
              write_timeout_min_seconds: 120

    Args:
        bucket (str): Name of bucket
        endpoint (str): URL of endpoint.
        access-key (str): S3-ACCESS-KEY
        secret-key (str): S3-SECRET-KEY
        read_timeout_seconds_per_kilobyte (float): S3 Read timeout in seconds/kilobyte
        write_timeout_seconds_per_kilobyte (float): S3 Write timeout in seconds/kilobyte
        read_timeout_min_seconds (float): The minimal timeout for S3 read
        write_timeout_min_seconds (float): The minimal timeout for S3 writes
        versioned_deletes (bool): Query and use the VersionId when performing deletes.
        hash-prefix-size (int): Number of hash characters to use as prefix in s3 object name.
        path-prefix-string (str): Additional string for path prefix
    """

    from botocore.config import Config as BotoConfig  # pylint: disable=import-outside-toplevel

    boto_config = BotoConfig(
        user_agent=S3_USERAGENT_NAME,
        connect_timeout=S3_TIMEOUT_CONNECT,
        read_timeout=S3_TIMEOUT_READ,
        retries={"max_attempts": S3_MAX_RETRIES},
        signature_version="s3v4",
    )

    return S3Storage(
        bucket,
        endpoint_url=endpoint,
        s3_read_timeout_seconds_per_kilobyte=read_timeout_seconds_per_kilobyte,
        s3_write_timeout_seconds_per_kilobyte=write_timeout_seconds_per_kilobyte,
        s3_read_timeout_min_seconds=read_timeout_min_seconds,
        s3_write_timeout_min_seconds=write_timeout_min_seconds,
        s3_versioned_deletes=versioned_deletes,
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key,
        config=boto_config,
        s3_hash_prefix_size=hash_prefix_size,
        s3_path_prefix_string=path_prefix_string,
    )


@object_tag("!redis-connection")
def load_redis_connection(
    host: Optional[str] = None,
    port: Optional[int] = None,
    password: Optional[str] = None,
    db: Optional[int] = None,
    dns_srv_record: Optional[str] = None,
    sentinel_master_name: Optional[str] = None,
    retries: int = 3,
) -> "RedisProvider":
    """Generates :class:`buildgrid.server.redis.provider.RedisProvider` using the tag ``!redis-connection``

    Usage
        .. code:: yaml

            - !redis-connection
              host: redis
              port: 6379
              password: !read-file /var/bgd/redis-pass
              db: 0
              dns-srv-record: <Domain name of SRV record>
              sentinel-master-name: <service_name of Redis sentinel's master instance>
              retries: 3


    Args:
        host (str | None): The hostname of the Redis server to use.
        port (int | None): The port that Redis is served on.
        password (str | None): The Redis database password to use.
        db (int): The Redis database number to use.
        dns-srv-record (str): Domain name of SRV record used to discover host/port
        sentinel-master-name (str): Service name of Redis master instance, used
            in a Redis sentinel configuration
        retries (int): Max number of times to retry (default 3). Backoff between retries is about 2^(N-1),
            where N is the number of attempts
    """

    # Import here so there is no global buildgrid dependency on redis
    from buildgrid.server.redis.provider import RedisProvider

    # ... validations like host/port xor dns srv record
    return RedisProvider(
        host=host,
        port=port,
        password=password,
        db=db,
        dns_srv_record=dns_srv_record,
        sentinel_master_name=sentinel_master_name,
        retries=retries,
    )


@object_tag("!redis-storage")
def load_redis_storage(redis: "RedisProvider") -> "RedisStorage":
    """Generates :class:`buildgrid.server.cas.storage.redis.RedisStorage` using the tag ``!redis-storage``.

    Usage
        .. code:: yaml

            # This assumes that the YAML anchors are defined elsewhere
            - !redis-storage
              redis: *redis-connection

    Args:
        redis (:class:`buildgrid.server.redis.provider.RedisProvider`): A configured Redis
            connection manager. This must be an object with an ``!redis-connection`` YAML tag.
    """

    # Import here so there is no global buildgrid dependency on redis
    from buildgrid.server.cas.storage.redis import RedisStorage

    return RedisStorage(redis)


@object_tag("!redis-index")
def load_redis_index(storage: StorageABC, redis: "RedisProvider") -> "RedisIndex":
    """Generates :class:`buildgrid.server.cas.storage.index.redis.RedisIndex`
    using the tag ``!redis-index``.

    Usage
        .. code:: yaml

            - !redis-index
              # This assumes that a storage instance is defined elsewhere
              # with a `&cas-storage` anchor
              storage: *cas-storage
              redis: *redis

    Args:
        storage(:class:`buildgrid.server.cas.storage.storage_abc.StorageABC`):
            Instance of storage to use. This must be a storage object constructed using
            a YAML tag ending in ``-storage``, for example ``!disk-storage``.
        redis (:class:`buildgrid.server.redis.provider.RedisProvider`): A configured Redis
            connection manager. This must be an object with an ``!redis-connection`` YAML tag.
    """

    # Import here so there is no global buildgrid dependency on redis
    from buildgrid.server.cas.storage.index.redis import RedisIndex

    return RedisIndex(redis=redis, storage=storage)


@object_tag("!replicated-storage")
def load_replicated_storage(storages: List[StorageABC]) -> ReplicatedStorage:
    """Generates :class:`buildgrid.server.cas.storage.replicated.ReplicatedStorage`
    using the tag ``!replicated-storage``.

    Usage
        .. code:: yaml

            - !replicated-storage
              storages:
                - &storageA
                - &storageB


    Args:
        Storages (list): List of storages to mirror reads/writes for.
            A minimum of two storages is required.
    """

    return ReplicatedStorage(storages)


@object_tag("!remote-storage")
def load_remote_storage(
    url: str,
    instance_name: Optional[str] = None,
    credentials: Optional[ClientCredentials] = None,
    channel_options: Optional[Dict[str, Any]] = None,
    retries: int = 3,
    max_backoff: int = 64,
    request_timeout: Optional[float] = None,
) -> RemoteStorage:
    """Generates :class:`buildgrid.server.cas.storage.remote.RemoteStorage`
    using the tag ``!remote-storage``.

    Usage
        .. code:: yaml

            - !remote-storage
              url: https://storage:50052/
              instance-name: main
              credentials:
                tls-server-key: !expand-path ~/.config/buildgrid/server.key
                tls-server-cert: !expand-path ~/.config/buildgrid/server.cert
                tls-client-certs: !expand-path ~/.config/buildgrid/client.cert
                auth-token: /path/to/auth/token
                token-refresh-seconds: 6000
              channel-options:
                lb-policy-name: round_robin
              request-timeout: 15


    Args:
        url (str): URL to remote storage. If used with ``https``, needs credentials.
        instance_name (str): Instance of the remote to connect to. If none, defaults to the instance context.
        credentials (dict, optional): A dictionary in the form::

           tls-client-key: /path/to/client-key
           tls-client-cert: /path/to/client-cert
           tls-server-cert: /path/to/server-cert
           auth-token: /path/to/auth/token
           token-refresh-seconds (int): seconds to wait before reading the token from the file again

        channel-options (dict, optional): A dictionary of grpc channel options in the form::

          some-channel-option: channel_value
          other-channel-option: another-channel-value
        See https://github.com/grpc/grpc/blob/master/include/grpc/impl/codegen/grpc_types.h
        for the valid channel options
        retries (int): Max number of times to retry (default 3). Backoff between retries is about 2^(N-1),
            where N is the number of attempts
        max_backoff (int): Maximum backoff in seconds (default 64)
        request_timeout (float): gRPC request timeout in seconds (default None)
    """

    options_tuple: Tuple[Tuple[str, Any], ...] = ()
    if channel_options:
        # Transform the channel options into the format expected
        # by grpc channel creation
        parsed_options = []
        for option_name, option_value in channel_options.items():
            parsed_options.append((f"grpc.{option_name.replace('-', '_')}", option_value))
        options_tuple = tuple(parsed_options)

    if not _validate_url_and_credentials(url, credentials=credentials):
        sys.exit(-1)

    return RemoteStorage(
        remote=url,
        instance_name=instance_name,
        channel_options=options_tuple,
        credentials=credentials,
        retries=retries,
        max_backoff=max_backoff,
        request_timeout=request_timeout,
    )


@object_tag("!with-cache-storage")
def load_with_cache_storage(
    cache: StorageABC,
    fallback: StorageABC,
    defer_fallback_writes: bool = False,
    fallback_writer_threads: int = 20,
) -> WithCacheStorage:
    """Generates :class:`buildgrid.server.cas.storage.with_cache.WithCacheStorage`
    using the tag ``!with-cache-storage``.

    Usage
        .. code:: yaml

            - !with-cache-storage
              cache:
                !lru-storage
                size: 2048M
              fallback:
                !disk-storage
                path: /opt/bgd/cas-storage
              defer-fallback-writes: no

    Args:
        cache (StorageABC): Storage instance to use as a cache
        fallback (StorageABC): Storage instance to use as a fallback on
            cache misses
        defer-fallback-writes (bool): If true, `commit_write` returns once
            writing to the cache is done, and the write into the fallback
            storage is done in a background thread
        fallback-writer-threads (int): The maximum number of threads to use
            for writing blobs into the fallback storage. Defaults to 20.
    """

    return WithCacheStorage(
        cache,
        fallback,
        defer_fallback_writes=defer_fallback_writes,
        fallback_writer_threads=fallback_writer_threads,
    )


class ShardType(TypedDict):
    name: str
    storage: StorageABC


@object_tag("!sharded-storage")
def load_sharded_storage(shards: List[ShardType], thread_pool_size: Optional[int] = None) -> ShardedStorage:
    """Generates :class:`buildgrid.server.cas.storage.Sharded.ShardedStorage`
    using the tag ``!sharded-storage``.

    Usage
        .. code:: yaml

            - !sharded-storage
              shards:
                - name: A
                  storage: &storageA
                - name: B
                  storage: !lru-storage
                    size: 2048M
              thread-pool-size: 40


    Args:
        shards (list): List of dictionaries. The dictionaries are expected to
            have ``name`` and ``storage`` keys defining a storage shard. The
            name must be unique within a configuration and should be the same
            for any configuration using the same underlying storage.
        thread-pool-size (int|None): Number of worker threads to use for bulk
            methods to allow parallel requests to each shard. If not set no
            threadpool is created and requests are made serially to each shard.
    """

    parsed_shards: Dict[str, StorageABC] = {}
    for shard in shards:
        if shard["name"] in parsed_shards:
            click.echo(
                f"ERROR: Duplicate shard name '{shard['name']}'. Please fix the config.\n",
                err=True,
            )
            sys.exit(-1)
        parsed_shards[shard["name"]] = shard["storage"]
    return ShardedStorage(parsed_shards, thread_pool_size)


_SizeLimitedStorageConfig = TypedDict("_SizeLimitedStorageConfig", {"max-size": int, "storage": StorageABC})


@object_tag("!size-differentiated-storage")
def load_size_differentiated_storage(
    size_limited_storages: List[_SizeLimitedStorageConfig],
    fallback: StorageABC,
    thread_pool_size: Optional[int] = None,
) -> SizeDifferentiatedStorage:
    """Generates :class:`buildgrid.server.cas.storage.size_differentiated.SizeDifferentiatedStorage`
    using the tag ``!size-differentiated-storage``.

    Usage
        .. code:: yaml

            - !size-differentiated-storage
              size-limited-storages:
                - max-size: 1M
                  storage:
                    !lru-storage
                    size: 2048M
              fallback:
                !disk-storage
                path: /opt/bgd/cas-storage
              thread-pool-size: 40

    Args:
        size_limited_storages (list): List of dictionaries. The dictionaries are expected
            to have ``max-size`` and ``storage`` keys, defining a storage provider to use
            to store blobs with size up to ``max-size``.
        fallback (StorageABC): Storage instance to use as a fallback for blobs which
            are too big for the options defined in ``size_limited_storages``.
        thread-pool-size (int|None): Number of worker threads to use for bulk
            methods to allow parallel requests to each storage. This thread pool
            is separate from the gRPC server thread-pool-size and should be tuned
            separately. If not set no threadpool is created and requests are made
            serially to each storage.
    """

    parsed_storages: List[SizeLimitedStorageType] = []
    for storage_config in size_limited_storages:
        parsed_storages.append(
            {"max_size": _parse_size(str(storage_config["max-size"])), "storage": storage_config["storage"]}
        )
    return SizeDifferentiatedStorage(parsed_storages, fallback, thread_pool_size)


@object_tag("!sql-storage")
def load_sql_storage(
    sql: SqlProvider,
    sql_ro: Optional[SqlProvider] = None,
) -> SQLStorage:
    """Generates :class:`buildgrid.server.cas.storage.sql.SQLStorage`
    using the tag ``!sql-storage``.

    Usage
        .. code:: yaml

            - !sql-storage
              sql: *sql
              sql_ro: *sql
    Args:
        sql (:class:`buildgrid.server.sql.provider.SqlProvider`): A configured SQL
            connection manager. This must be an object with an ``!sql-connection`` YAML tag.
        sql_ro (:class:`buildgrid.server.sql.provider.SqlProvider`): Similar to `sql`,
            but used for readonly backend transactions.
            If set, it should be configured with a replica of main DB using an optional but
            encouraged readonly role. Permission check is not executed by BuildGrid.
            If not set, readonly transactions are executed by `sql` object.
    """

    return SQLStorage(sql, sql_ro_provider=sql_ro)


@object_tag("!sql-connection")
def load_sql_connection(
    automigrate: bool = False,
    connection_string: Optional[str] = None,
    connection_timeout: int = 5,
    lock_timeout: int = 5,
    connect_args: Optional[Dict[str, Any]] = None,
    max_overflow: Optional[int] = None,
    pool_pre_ping: Optional[bool] = None,
    pool_recycle: Optional[int] = None,
    pool_size: Optional[int] = None,
    pool_timeout: Optional[int] = None,
) -> SqlProvider:
    """Generates :class:`buildgrid.server.sql.provider.SqlProvider` using the
    tag ``!sql-connection``.

    Example:
        .. code:: yaml

            - !sql-connection &sql
              connection_string: postgresql://bgd:insecure@database/bgd
              automigrate: yes
              connection_timeout: 5
              lock_timeout: 5
              pool-size: 5
              pool-timeout: 30
              max-overflow: 10

    """

    return SqlProvider(
        automigrate=automigrate,
        connection_string=connection_string,
        connection_timeout=connection_timeout,
        lock_timeout=lock_timeout,
        connect_args=connect_args,
        max_overflow=max_overflow,
        pool_pre_ping=pool_pre_ping,
        pool_recycle=pool_recycle,
        pool_size=pool_size,
        pool_timeout=pool_timeout,
    )


@object_tag(
    "!sql-scheduler",
    deprecated_keys={
        "property-keys": "Use a 'property-set' object instead. See '!dynamic-property-set'.",
        "wildcard-property-keys": "Use a 'property-set' object instead. See '!dynamic-property-set'.",
    },
)
def load_sql_scheduler(
    storage: StorageABC,
    sql: SqlProvider,
    sql_ro: Optional[SqlProvider] = None,
    sql_notifier: Optional[SqlProvider] = None,
    property_set: Optional[PropertySet] = None,
    pruner_job_max_age: Optional[Dict[str, float]] = None,
    pruner_period: Optional[Dict[str, float]] = None,
    pruner_max_delete_window: Optional[int] = None,
    queue_timeout_job_max_age: Optional[Dict[str, float]] = None,
    queue_timeout_period: Optional[Dict[str, float]] = None,
    queue_timeout_max_window: Optional[int] = None,
    action_cache: Optional[ActionCacheABC] = None,
    action_browser_url: Optional[str] = None,
    max_execution_timeout: int = DEFAULT_MAX_EXECUTION_TIMEOUT,
    metering_service_client: Optional[metering.SyncMeteringServiceClient] = None,
    metering_throttle_action: Optional[str] = None,
    bot_session_keepalive_timeout: int = 600,
    logstream: Optional[Dict[str, Any]] = None,
    asset_client: Optional[AssetClient] = None,
    queued_action_retention_hours: Optional[float] = None,
    completed_action_retention_hours: Optional[float] = None,
    action_result_retention_hours: Optional[float] = None,
    max_job_attempts: int = 5,
    priority_assignment_percentage: int = 100,
    poll_interval: float = 1.0,
    max_queue_size: Optional[int] = None,
    # Deprecated values:
    property_keys: Optional[Union[str, List[str]]] = None,
    wildcard_property_keys: Optional[Union[str, List[str]]] = None,
) -> Scheduler:
    """Generates :class:`buildgrid.server.scheduler.Scheduler` using
    the tag ``!sql-scheduler``.

    Example:

        .. code:: yaml

            - !sql-scheduler
              storage: *cas-storage
              sql: *sql
              pruner-job-max-age:
                days: 90

        This usage example assumes that the ``cas-storage`` reference refers to a
        storage backend, eg. ``!disk-storage``, and the ``sql`` reference refers
        to an SQL connection manager using ``!sql-connection``.

    Args:

        storage(:class:`buildgrid.server.cas.storage.storage_abc.StorageABC`): Instance
            of storage to use for getting actions and storing job results. This must be
            an object constructed using a YAML tag ending in ``-storage``, for example
            ``!disk-storage``.

        sql (:class:`buildgrid.server.sql.provider.SqlProvider`): A configured SQL
            connection manager. This must be an object with an ``!sql-connection`` YAML tag.

        sql_ro (:class:`buildgrid.server.sql.provider.SqlProvider`): Similar to `sql`,
            but used for readonly backend transactions.
            If set, it should be configured with a replica of main DB using an optional but
            encouraged readonly role. Permission check is not executed by BuildGrid.
            If not set, readonly transactions are executed by `sql` object.

        sql_notifier (:class:`buildgrid.server.sql.provider.SqlProvider`): Similar to `sql`,
            but used for operation notifier.
            If not set, transactions are executed by `sql` object.

        property_set (PropertySet): Controls how execute requests are assigned to workers.

        pruner_job_max_age (dict): Allow the storage to remove old entries by specifying the
            maximum amount of time that a row should be kept after its job finished. If
            this value is None, pruning is disabled and the background pruning thread
            is never created.

        pruner_period (dict): How often to attempt to remove old entries. If pruning
            is enabled (see above) and this value is None, it is set to 5 minutes by default.

        pruner_max_delete_window (int): Maximum number of records removed in a single
            cleanup pass. If pruning is enabled and this value is None, it is set to 10000
            by default. This allows to put a limit on the time that the database
            will be blocked on a single invocation of the cleanup routine.
            (A smaller value reduces the performance impact of removing entries,
            but makes the recovery of storage space slower.)

        queue_timeout_job_max_age (dict): If set, allow storage to abort jobs that have been queued
            for a long period of time.

        queue_timeout_period (dict): How often to find aged queued jobs. If not set,
            default to 5 minutes.

        queue_timeout_max_window (int): Maximum number of jobs to timeout per batch.
            If not set, default to 10000.

        action_cache (:class:`ActionCache`): Instance of action cache to use.

        action_browser_url (str): The base URL to use to generate Action Browser links to users.
            If a single Web interface serves several Buildgrid installations then this URL
            should include the namespace configured for the current Buildgrid installation,
            see https://gitlab.com/BuildGrid/bgd-browser#multi-buildgrid-setup.

        max_execution_timeout (int): The maximum time jobs are allowed to be in
            'OperationStage.EXECUTING'. This is a periodic check.
            When this time is exceeded in executing stage, the job will be cancelled.

        metering_service_client: Optional client to check whether resource usage of a client
            is above a predefined threshold

        metering_throttle_action: The action to perform when metering service returns that job should
            be throttled. Can be set to "deprioritize" or "reject". Defaults to "deprioritize".

        bot_session_keepalive_timeout (int): The longest time (in seconds) we'll wait
            for a bot to send an update before it assumes it's dead. Defaults to 600s
            (10 minutes).

        logstream (Dict): Configuration options for connecting a logstream instance to ongoing
            jobs. Is a dict with items "url", "credentials", and "instance-name"

        asset_client (Optional[AssetClient]): Client of remote-asset service

        queued_action_retention_hours (Optional[float]): Minimum retention for queued actions in hours

        completed_action_retention_hours (Optional[float]): Minimum retention for completed actions in hours

        action_result_retention_hours (Optional[float]): Minimum retention for action results in hours

        max_job_attempts (int): The number of times a job will be assigned to workers before marking
            the job failed. Reassignment happens when a worker fails to report the outcome of a job.
            Minimum value allowed is 1. Default value is 5.

        priority_assignment_percentage (int): A value between 0 and 100 (inclusive) representing
            the percentage of workers to assign jobs to in priority order. The remainder will be
            assigned work in oldest-first order. Defaults to 100, or all work assigned in priority
            order.

        poll_interval (float): Duration to wait between periodic operations.

        max_queue_size (int): Maximum number of jobs queued per platform property set.

        property_keys (list): Deprecated. Use a property_set instead.
        wildcard_property_keys (list): Deprecated. Use a property_set instead.
    """

    click.echo(
        f"SQLScheduler: storage={type(storage).__name__}, "
        f"pruner_job_max_age={pruner_job_max_age}, "
        f"pruner_period={pruner_period}, "
        f"pruner_max_delete_window={pruner_max_delete_window}"
    )
    click.echo(click.style("Creating an SQL scheduler backend\n", fg="green", bold=True))

    if bot_session_keepalive_timeout <= 0:
        msg = f"ERROR: bot_session_keepalive_timeout must be greater than zero: {bot_session_keepalive_timeout}"
        click.echo(click.style(msg, fg="red", bold=True), err=True)
        sys.exit(-1)

    if max_job_attempts < 1:
        msg = f"ERROR: max_job_attempts must be greater than zero: {max_job_attempts}"
        click.echo(click.style(msg, fg="red", bold=True), err=True)
        sys.exit(-1)

    try:
        if property_set is None:
            if isinstance(property_keys, str):
                property_keys = [property_keys]
            if isinstance(wildcard_property_keys, str):
                wildcard_property_keys = [wildcard_property_keys]

            property_set = load_dynamic_property_set(
                match_property_keys=property_keys,
                wildcard_property_keys=wildcard_property_keys,
            )

        pruning_options = (
            AgedJobHandlerOptions.from_config(pruner_job_max_age, pruner_period, pruner_max_delete_window)
            if pruner_job_max_age
            else None
        )
        queue_timeout_options = (
            AgedJobHandlerOptions.from_config(
                queue_timeout_job_max_age, queue_timeout_period, queue_timeout_max_window
            )
            if queue_timeout_job_max_age
            else None
        )

        sql_ro = sql_ro or sql
        sql_notifier = sql_notifier or sql

        logstream_url, logstream_credentials, logstream_instance = get_logstream_connection_info(logstream)
        logstream_channel: Optional[grpc.Channel] = None
        if logstream_url is not None:
            logstream_credentials = logstream_credentials or {}
            logstream_channel, _ = setup_channel(
                logstream_url,
                auth_token=None,
                client_key=logstream_credentials.get("tls-client-key"),
                client_cert=logstream_credentials.get("tls-client-cert"),
                server_cert=logstream_credentials.get("tls-server-cert"),
            )

        return Scheduler(
            sql,
            storage,
            property_set=property_set,
            pruning_options=pruning_options,
            queue_timeout_options=queue_timeout_options,
            sql_ro_provider=sql_ro,
            sql_notifier_provider=sql_notifier,
            action_cache=action_cache,
            action_browser_url=action_browser_url,
            max_execution_timeout=max_execution_timeout,
            metering_client=metering_service_client,
            metering_throttle_action=(
                MeteringThrottleAction(metering_throttle_action)
                if metering_throttle_action
                else MeteringThrottleAction.DEPRIORITIZE
            ),
            bot_session_keepalive_timeout=bot_session_keepalive_timeout,
            logstream_channel=logstream_channel,
            logstream_instance=logstream_instance,
            asset_client=asset_client,
            queued_action_retention_hours=queued_action_retention_hours,
            completed_action_retention_hours=completed_action_retention_hours,
            action_result_retention_hours=action_result_retention_hours,
            priority_assignment_percentage=priority_assignment_percentage,
            poll_interval=poll_interval,
            max_queue_size=max_queue_size,
        )

    except TypeError as type_error:
        click.echo(type_error, err=True)
        sys.exit(-1)


@object_tag("!dynamic-property-set")
def load_dynamic_property_set(
    unique_property_keys: Optional[Iterable[str]] = None,
    match_property_keys: Optional[Iterable[str]] = None,
    wildcard_property_keys: Optional[Iterable[str]] = None,
    label_key: Optional[str] = None,
) -> DynamicPropertySet:
    """
    A dynamic property set allows scheduling jobs which may have unset values for properties.
    Dynamic queues can be flexible as they allow minimal configuration to add new properties,
    however, they have an exponential cost to scheduling. Using many different properties
    can lead to very slow scheduling rates.

    Args:
        unique_property_keys(Set[str]): Properties which may only be set once.
            OSFamily is always considered unique.
        match_property_keys(Set[str]): Properties which must match on the worker and execute request.
            OSFamily and ISA property keys are always added as match keys even if unlisted.
        wildcard_property_keys(Set[str]): Properties which are available to workers, but not used for scheduling.
        label_key(str): A key used to identify job types in logging and metrics.
            Defaults to OSFamily
    """

    match_property_keys = set(match_property_keys or [])
    match_property_keys.update(DEFAULT_PLATFORM_PROPERTY_KEYS)

    unique_property_keys = set(unique_property_keys or [])
    unique_property_keys.add("OSFamily")

    wildcard_property_keys = set(wildcard_property_keys or [])

    label_key = label_key or "OSFamily"

    return DynamicPropertySet(
        unique_property_keys=unique_property_keys,
        match_property_keys=match_property_keys,
        wildcard_property_keys=wildcard_property_keys,
        label_key=label_key,
    )


_Properties = List[Tuple[str, str]]


class _PropertyLabel(TypedDict):
    label: str
    properties: _Properties


@object_tag("!static-property-set")
def load_static_property_set(
    property_labels: List[_PropertyLabel],
    wildcard_property_keys: Optional[List[str]] = None,
) -> StaticPropertySet:
    """
    A static property set allows scheduling jobs by resolving sane defaults for unspecified keys.
    Static queues can be more verbose as you require defining all sets of valid properties,
    however, they have a linear cost to scheduling. Using many different properties
    becomes less expensive to calculate assignment.

    Args:
        property_labels(List[_PropertyLabel]): Properties combinations which are allowed.
        wildcard_property_keys(List[str]): Properties which are available to workers, but not used for scheduling.
    """

    if wildcard_property_keys is None:
        wildcard_property_keys = []

    return StaticPropertySet(
        property_labels=[
            PropertyLabel(
                label=label["label"],
                properties={(k, v) for [k, v] in label["properties"]},
            )
            for label in property_labels
        ],
        wildcard_property_keys=set(wildcard_property_keys),
    )


@object_tag("!sql-index")
def load_sql_index(
    storage: StorageABC,
    sql: SqlProvider,
    window_size: int = 1000,
    inclause_limit: int = -1,
    fallback_on_get: bool = False,
    max_inline_blob_size: int = 0,
    refresh_accesstime_older_than: int = 0,
) -> SQLIndex:
    """Generates :class:`buildgrid.server.cas.storage.index.sql.SQLIndex`
    using the tag ``!sql-index``.

    Usage
        .. code:: yaml

            - !sql-index
              # This assumes that a storage instance is defined elsewhere
              # with a `&cas-storage` anchor
              storage: *cas-storage
              sql: *sql
              window-size: 1000
              inclause-limit: -1
              fallback-on-get: no
              max-inline-blob-size: 256
              refresh-accesstime-older-than: 0

    Args:
        storage(:class:`buildgrid.server.cas.storage.storage_abc.StorageABC`):
            Instance of storage to use. This must be a storage object constructed using
            a YAML tag ending in ``-storage``, for example ``!disk-storage``.
        window_size (uint): Maximum number of blobs to fetch in one SQL operation
            (larger resultsets will be automatically split into multiple queries)
        inclause_limit (int): If nonnegative, overrides the default number of variables
            permitted per "in" clause. See the buildgrid.server.cas.storage.index.sql.SQLIndex
            comments for more details.
        fallback_on_get (bool): By default, the SQL Index only fetches blobs from the
            underlying storage if they're present in the index on ``get_blob``/``bulk_read_blobs``
            requests to minimize interactions with the storage. If this is set, the index
            instead checks the underlying storage directly on ``get_blob``/``bulk_read_blobs``
            requests, then loads all blobs found into the index.
        max_inline_blob_size (int): Blobs of this size or smaller are stored directly in the index
            and not in the backing storage (must be nonnegative).
        refresh-accesstime-older-than (int): When reading a blob, its access timestamp will not be
            updated if the current time is not at least refresh-accesstime-older-than seconds newer
            than the access timestamp. Set this to reduce load associated with frequent timestamp updates.
    """

    storage_type = type(storage).__name__
    click.echo(
        f"SQLIndex: storage={storage_type}, "
        f"window_size={window_size}, "
        f"inclause_limit={inclause_limit}, "
        f"fallback_on_get={fallback_on_get}"
    )
    click.echo(click.style(f"Creating an SQL CAS Index for {storage_type}\n", fg="green", bold=True))
    return SQLIndex(
        sql,
        storage,
        window_size=window_size,
        inclause_limit=inclause_limit,
        fallback_on_get=fallback_on_get,
        max_inline_blob_size=max_inline_blob_size,
        refresh_accesstime_older_than=refresh_accesstime_older_than,
    )


@object_tag("!execution")
def load_execution_controller(
    scheduler: Scheduler,
    operation_stream_keepalive_timeout: int = 600,
    endpoints: Sequence[str] = ServiceName.default_services(),
    max_list_operations_page_size: int = DEFAULT_MAX_LIST_OPERATION_PAGE_SIZE,
) -> ExecutionController:
    """Generates :class:`buildgrid.server.execution.service.ExecutionService`
    using the tag ``!execution``.

    Usage
        .. code:: yaml

            # This assumes that the YAML anchors are defined elsewhere
            - !execution
              scheduler: *state-database
              operation-stream-keepalive-timeout: 600
              endpoints:
                - execution
                - operations
                - bots
              max-list-operations-page-size: 1000

    Args:
        scheduler(:class:`Scheduler`): Instance of scheduler to use for the scheduler's state.
        operation_stream_keepalive_timeout (int): The longest time (in seconds)
            we'll wait before sending the current status in an Operation response
            stream of an `Execute` or `WaitExecution` request. Defaults to 600s
            (10 minutes).
        endpoints (list): List of service/endpoint types to enable. Possible services are
            ``execution``, ``operations``, and ``bots``. By default all three are enabled.
        max_list_operations_page_size (int): The maximum number of operations that can
            be returned in a ListOperations response. A page token will be returned
            with the response to allow the client to get the next page of results.
    """

    click.echo(click.style(f"Creating a Execution service using {type(scheduler).__name__}\n", fg="green", bold=True))

    return ExecutionController(
        scheduler,
        operation_stream_keepalive_timeout=operation_stream_keepalive_timeout,
        services=endpoints,
        max_list_operations_page_size=max_list_operations_page_size,
    )


@object_tag("!bots")
def load_bots_controller(scheduler: Scheduler) -> BotsInterface:
    """Generates :class:`buildgrid.server.bots.instance.BotsInterface`
    using the tag ``!bots``.

    Usage
        .. code:: yaml

            # This assumes that the YAML anchors are defined elsewhere
            - !bots
              scheduler: *state-database

    Args:
        scheduler(:class:`Scheduler`): Instance of scheduler to use for the scheduler's state.
    """

    click.echo(click.style(f"Creating a Bots service using {type(scheduler).__name__}\n", fg="green", bold=True))
    return BotsInterface(scheduler)


@object_tag("!action-cache")
def load_action_cache_controller(cache: ActionCacheABC) -> ActionCache:
    """Generates :class:`buildgrid.server.actioncache.service.ActionCacheService`
    using the tag ``!action-cache``.

    Usage
        .. code:: yaml

            # This assumes that the YAML anchors are defined elsewhere
            - !action-cache
              cache: *lru-cache

    Args:
        cache (ActionCacheABC): The ActionCache backend to use for this cache.

    """

    click.echo(click.style(f"Creating a Action Cache service using {type(cache).__name__}\n", fg="green", bold=True))
    return ActionCache(cache)


@object_tag("!mirrored-action-cache")
def load_mirrored_action_cache(first: ActionCacheABC, second: ActionCacheABC) -> MirroredCache:
    """Generates:class:`buildgrid.server.actioncache.caches.mirrored_cache.MirroredCache`
    using the tag ``!mirrored-action-cache``.

    Usage
        .. code:: yaml

            # This assumes that the YAML anchors are defined elsewhere
            - !mirrored-action-cache
              first: *first-action-cache
              second: *second-action-cache
    """

    return MirroredCache(first=first, second=second)


@object_tag("!with-cache-action-cache")
def load_with_cache_action_cache(
    cache: ActionCacheABC,
    fallback: ActionCacheABC,
    allow_updates: bool = True,
    cache_failed_actions: bool = True,
) -> WithCacheActionCache:
    """Generates:class:`buildgrid.server.actioncache.caches.with_cache.WithCacheActionCache`
    using the tag ``!with-cache-action-cache``.

    Usage
        .. code:: yaml

            # This assumes that the YAML anchors are defined elsewhere
            - !with-cache-action-cache
              storage: *cas-storage
              cache: *cache-ac
              fallback: *fallback-ac

    Args:
        cache (ActionCacheABC): ActionCache instance to use as a local cache
        fallback (ActionCacheABC): ActionCache instance to use as a fallback on
            local cache misses
        allow_updates(bool): Allow updates pushed to the Action Cache.
            Defaults to ``True``.
        cache_failed_actions(bool): Whether to store failed (non-zero exit
            code) actions. Default to ``True``.
    """

    return WithCacheActionCache(
        cache, fallback, allow_updates=allow_updates, cache_failed_actions=cache_failed_actions
    )


@object_tag("!lru-action-cache")
def load_lru_action_cache(
    storage: StorageABC,
    max_cached_refs: int,
    allow_updates: bool = True,
    cache_failed_actions: bool = True,
) -> LruActionCache:
    """Generates :class:`buildgrid.server.actioncache.caches.lru_cache.LruActionCache`
    using the tag ``!lru-action-cache``.

    Usage
        .. code:: yaml

            # This assumes that the YAML anchors are defined elsewhere
            - !lru-action-cache
              storage: *cas-storage
              max-cached-refs: 1024
              cache-failed-actions: yes
              allow-updates: yes

    Args:
        storage(:class:`buildgrid.server.cas.storage.storage_abc.StorageABC`):
            Instance of storage to use.
        max_cached_refs(int): Max number of cached actions.
        allow_updates(bool): Allow updates pushed to the Action Cache.
            Defaults to ``True``.
        cache_failed_actions(bool): Whether to store failed (non-zero exit
            code) actions. Default to ``True``.

    """

    storage_type = type(storage).__name__
    click.echo(
        f"LruActionCache: storage={storage_type}, max_cached_refs={max_cached_refs}, "
        f"allow_updates={allow_updates}, cache_failed_actions={cache_failed_actions}"
    )
    click.echo(click.style(f"Creating an LruActionCache using `{storage_type}` storage\n", fg="green", bold=True))
    return LruActionCache(storage, max_cached_refs, allow_updates, cache_failed_actions)


@object_tag("!s3action-cache")
def load_s3_action_cache(
    storage: StorageABC,
    allow_updates: bool = True,
    cache_failed_actions: bool = True,
    entry_type: Optional[str] = None,
    migrate_entries: Optional[bool] = False,
    bucket: Optional[str] = None,
    endpoint: Optional[str] = None,
    access_key: Optional[str] = None,
    secret_key: Optional[str] = None,
) -> S3ActionCache:
    """Generates :class:`buildgrid.server.actioncache.caches.s3_cache.S3ActionCache`
    using the tag ``!s3action-cache``.

    Usage
        .. code:: yaml

            # This assumes that the YAML anchors are defined elsewhere
            - !s3action-cache
              storage: *cas-storage
              allow-updates: yes
              cache-failed-actions: yes
              entry-type: action-result-digest
              migrate-entries: no
              bucket: bgd-action-cache
              endpoint: http://localhost:9000/
              access-key: !read-file /var/bgd/s3-access-key
              secret-key: !read-file /var/bgd/s3-secret-key

    Args:
        storage(:class:`buildgrid.server.cas.storage.storage_abc.StorageABC`):
            Instance of storage to use. This must be an object constructed using
            a YAML tag ending in ``-storage``, for example ``!disk-storage``.
        allow_updates(bool): Allow updates pushed to the Action Cache.
            Defaults to ``True``.
        cache_failed_actions(bool): Whether to store failed (non-zero exit code)
            actions. Default to ``True``.
        entry_type (str): whether entries in S3 will store an ``'action-result'``
            or an ``'action-result-digest'`` (default).
        migrate_entries (bool): Whether to automatically update the values of
            entries that contain a different type of value to `entry_type` as
            they are queried. Default to ``False``.
        bucket (str): Name of bucket
        endpoint (str): URL of endpoint.
        access-key (str): S3-ACCESS-KEY
        secret-key (str): S3-SECRET-KEY

    """

    storage_type = type(storage).__name__
    cache_entry_type = None

    if entry_type is None or entry_type.lower() == "action-result-digest":
        cache_entry_type = ActionCacheEntryType.ACTION_RESULT_DIGEST
    elif entry_type.lower() == "action-result":
        cache_entry_type = ActionCacheEntryType.ACTION_RESULT
    else:
        click.echo(
            click.style(f"ERROR: entry_type value is not valid: {cache_entry_type}", fg="red", bold=True), err=True
        )
        sys.exit(-1)

    click.echo(
        f"S3ActionCache: storage={storage_type}, allow_updates={allow_updates}, "
        f"cache_failed_actions={cache_failed_actions}, bucket={bucket}, "
        f"entry_type={entry_type}, migrate_entries={migrate_entries}, "
        f"endpoint={endpoint}"
    )
    click.echo(
        click.style(f"Creating an S3ActionCache service using `{storage_type}` storage\n", fg="green", bold=True)
    )

    from botocore.config import Config as BotoConfig  # pylint: disable=import-outside-toplevel

    boto_config = BotoConfig(
        user_agent=S3_USERAGENT_NAME,
        connect_timeout=S3_TIMEOUT_CONNECT,
        read_timeout=S3_TIMEOUT_READ,
        retries={"max_attempts": S3_MAX_RETRIES},
    )

    return S3ActionCache(
        storage,
        allow_updates=allow_updates,
        cache_failed_actions=cache_failed_actions,
        entry_type=cache_entry_type,
        migrate_entries=migrate_entries,
        bucket=bucket,
        endpoint_url=endpoint,
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key,
        config=boto_config,
    )


@object_tag("!remote-action-cache")
def load_remote_action_cache(
    url: str,
    instance_name: Optional[str] = None,
    retries: int = 3,
    max_backoff: int = 64,
    request_timeout: Optional[float] = None,
    credentials: Optional[ClientCredentials] = None,
    channel_options: Optional[Dict[str, Any]] = None,
) -> RemoteActionCache:
    """Generates :class:`buildgrid.server.actioncache.caches.remote.RemoteActionCache`
    using the tag ``!remote-action-cache``.

    Usage
        .. code:: yaml

            - !remote-action-cache
              url: https://action-cache:50053
              instance-name: main
              credentials:
                tls-server-key: !expand-path ~/.config/buildgrid/server.key
                tls-server-cert: !expand-path ~/.config/buildgrid/server.cert
                tls-client-certs: !expand-path ~/.config/buildgrid/client.cert
                auth-token: /path/to/auth/token
                token-refresh-seconds: 6000
              channel-options:
                lb-policy-name: round_robin

    Args:
        url (str): URL to remote action cache
        instance_name (Optional[str]): Instance of the remote to connect to.
            Defaults to the instance context if none.
        credentials (dict, optional): A dictionary in the form::

           tls-client-key: /path/to/client-key
           tls-client-cert: /path/to/client-cert
           tls-server-cert: /path/to/server-cert
           auth-token: /path/to/auth/token
           token-refresh-seconds (int): seconds to wait before reading the token from the file again

        channel-options (dict, optional): A dictionary of grpc channel options in the form::

          some-channel-option: channel_value
          other-channel-option: another-channel-value
        See https://github.com/grpc/grpc/blob/master/include/grpc/impl/codegen/grpc_types.h
        for the valid channel options

    """

    options_tuple: Tuple[Tuple[str, Any], ...] = ()
    if channel_options:
        # Transform the channel options into the format expected
        # by grpc channel creation
        parsed_options = []
        for option_name, option_value in channel_options.items():
            parsed_options.append((f"grpc.{option_name.replace('-', '_')}", option_value))
        options_tuple = tuple(parsed_options)

    if not _validate_url_and_credentials(url, credentials=credentials):
        sys.exit(-1)

    click.echo(f"RemoteActionCache: url={url}, instance_name={instance_name}, ")
    click.echo(click.style(f"Creating an RemoteActionCache service for {url}\n", fg="green", bold=True))

    return RemoteActionCache(
        url,
        instance_name,
        retries,
        max_backoff,
        request_timeout,
        channel_options=options_tuple,
        credentials=credentials,
    )


@object_tag("!write-once-action-cache")
def load_write_once_action_cache(action_cache: ActionCacheABC) -> WriteOnceActionCache:
    """Generates :class:`buildgrid.server.actioncache.caches.write_once_cache.WriteOnceActionCache`
    using the tag ``!write-once-action-cache``.

    This allows a single update for a given key, essentially making it possible
    to create immutable ActionCache entries, rather than making the cache read-only
    as the ``allow-updates`` property of other ActionCache implementations does.

    Usage
        .. code:: yaml

            # This assumes that the YAML anchors are defined elsewhere
            - !write-once-action-cache
              action-cache: *remote-cache

    Args:
        action_cache (ActionCache): The action cache instance to make immutable.

    """

    return WriteOnceActionCache(action_cache)


@object_tag("!redis-action-cache")
def load_redis_action_cache(
    storage: StorageABC,
    # Should be RedisProvider, but we are trying to avoid a global dependency on redis.
    # Parser should have validated this already and we assert below
    redis: Any,
    allow_updates: bool = True,
    cache_failed_actions: bool = True,
    entry_type: Optional[str] = None,
    migrate_entries: Optional[bool] = False,
) -> "RedisActionCache":
    """Generates :class:`buildgrid.server.actioncache.caches.redis_cache.RedisActionCache`
    using the tag ``!redis-action-cache``.

    This creates an Action Cache which stores the mapping from Action digests to
    ActionResults in Redis.

    Usage
        .. code:: yaml

            # This assumes that the YAML anchors are defined elsewhere
            - !redis-action-cache
              storage: *cas-storage
              allow-updates: yes
              cache-failed-actions: yes
              entry-type: action-result-digest
              migrate-entries: no
              redis: *redis-connection

    Args:
        storage(:class:`buildgrid.server.cas.storage.storage_abc.StorageABC`):
            Instance of storage to use. This must be an object constructed using
            a YAML tag ending in ``-storage``, for example ``!disk-storage``.
        allow_updates(bool): Allow updates pushed to the Action Cache.
            Defaults to ``True``.
        cache_failed_actions(bool): Whether to store failed (non-zero exit code)
            actions. Default to ``True``.
        entry_type (str): whether entries in Redis will store an ``'action-result'``
            or an ``'action-result-digest'`` (default).
        migrate_entries (bool): Whether to automatically update the values of
            entries that contain a different type of value to `entry_type` as
            they are queried. Default to ``False``.
        redis (:class:`buildgrid.server.redis.provider.RedisProvider`): A configured Redis
            connection manager. This must be an object with an ``!redis-connection`` YAML tag.
    """

    cache_entry_type = None
    if entry_type is None or entry_type.lower() == "action-result-digest":
        cache_entry_type = ActionCacheEntryType.ACTION_RESULT_DIGEST
    elif entry_type.lower() == "action-result":
        cache_entry_type = ActionCacheEntryType.ACTION_RESULT
    else:
        click.echo(
            click.style(f"ERROR: entry_type value is not valid: {cache_entry_type}", fg="red", bold=True), err=True
        )
        sys.exit(-1)
    # Import here so there is no global buildgrid dependency on redis
    from buildgrid.server.actioncache.caches.redis_cache import RedisActionCache
    from buildgrid.server.redis.provider import RedisProvider

    try:
        assert isinstance(redis, RedisProvider)
        return RedisActionCache(
            storage,
            redis,
            allow_updates=allow_updates,
            cache_failed_actions=cache_failed_actions,
            entry_type=cache_entry_type,
            migrate_entries=migrate_entries,
        )
    except Exception as e:
        click.echo(click.style(f"ERROR: {e},", fg="red", bold=True), err=True)
        sys.exit(-1)


@object_tag("!cas")
def load_cas_controller(
    storage: StorageABC,
    read_only: bool = False,
    tree_cache_size: Optional[int] = None,
    tree_cache_ttl_minutes: float = 60,
) -> ContentAddressableStorageInstance:
    """Generates :class:`buildgrid.server.cas.service.ContentAddressableStorageService`
    using the tag ``!cas``.

    Usage
        .. code:: yaml

            # This assumes that the YAML anchors are defined elsewhere
            - !cas
              storage: *cas-storage

    Args:
        storage(:class:`buildgrid.server.cas.storage.storage_abc.StorageABC`):
            Instance of storage to use. This must be an object constructed using
            a YAML tag ending in ``-storage``, for example ``!disk-storage``.

        tree_cache_size (Optional[int]): Size of GetTreeResponse cache, default to None.
            This feature is experimental for testing purposes.
            It could be deprecated in favor of a redis cache in future.

        tree_cache_ttl_minutes (float): TTL of GetTreeResponse cache, default to 60 minutes.
            This feature is experimental for testing purposes.
            It could be deprecated in favor of a redis cache in future.
    """

    click.echo(click.style(f"Creating a CAS service using {type(storage).__name__}\n", fg="green", bold=True))
    return ContentAddressableStorageInstance(
        storage,
        read_only=read_only,
        tree_cache_size=tree_cache_size,
        tree_cache_ttl_minutes=tree_cache_ttl_minutes,
    )


@object_tag("!bytestream")
def load_bytestream_controller(
    storage: StorageABC,
    read_only: bool = False,
    disable_overwrite_early_return: bool = False,
) -> ByteStreamInstance:
    """Generates :class:`buildgrid.server.cas.service.ByteStreamService`
    using the tag ``!bytestream``.

    Usage
        .. code:: yaml

            # This assumes that the YAML anchors are defined elsewhere
            - !bytestream
              storage: *cas-storage

    Args:
        storage(:class:`buildgrid.server.cas.storage.storage_abc.StorageABC`):
            Instance of storage to use. This must be an object constructed using
            a YAML tag ending in ``-storage``, for example ``!disk-storage``.
    """

    click.echo(
        click.style(f"Creating a ByteStream service using storage {type(storage).__name__}", fg="green", bold=True)
    )
    return ByteStreamInstance(
        storage,
        read_only=read_only,
        disable_overwrite_early_return=disable_overwrite_early_return,
    )


@object_tag("!memory-build-events")
def load_memory_build_events_controller() -> BuildEventStreamStorage:
    """Generates :class:`buildgrid.server.build_events.storage.BuildEventStreamStorage`
    using the tag ``!memory-build-events-storage``.

    Usage
        .. code:: yaml

            - !memory-build-events
    """

    return BuildEventStreamStorage()


@object_tag("!metering-service-client")
def load_metering_service_client(
    base_url: str,
    token_path: Optional[str] = None,
    retry_max_attempts: int = 0,  # Default to no retry
    retry_exp_base: float = 1.5,
    retry_multiplier: float = 1.0,
    retry_max_wait: float = 10.0,
    retry_http_statuses: Optional[List[int]] = None,
    retry_exceptions: Optional[List[str]] = None,
    retry_cause_exceptions: Optional[List[str]] = None,
) -> metering.SyncMeteringServiceClient:
    """Generates :class:`buildgrid_metering.client.SyncMeteringServiceClient`
    using the tag ``!metering-service-client``.

    Usage
        .. code:: yaml

            - !metering-service-client
              token-path: /tmp/path/to/token
              retry-max-attempts: 3
              retry-exp-base: 2
              retry-multiplier: 1
              retry-http-statuses: [503]
              retry-exceptions: ["metering-service-client-error"]
    """

    if token_path is not None:
        auth_config = metering.auth.AuthTokenConfig(mode=metering.auth.AuthTokenMode.FILEPATH, value=token_path)
    else:
        auth_config = metering.auth.AuthTokenConfig(mode=metering.auth.AuthTokenMode.NONE, value="")

    def _get_exception_class(name: str) -> Type[Exception]:
        exception_classes = {
            "metering-service-error": MeteringServiceError,
            "metering-service-client-error": MeteringServiceClientError,
            "timeout-error": requests.ConnectionError,
        }
        try:
            return exception_classes[name]
        except KeyError:
            raise ValueError(f"Unsupported exception class: {name}. Supported classes: {exception_classes.keys()}")

    retry_config = metering.RetryConfig(
        max_attempts=retry_max_attempts,
        exp_base=retry_exp_base,
        multiplier=retry_multiplier,
        max_wait=retry_max_wait,
        http_statuses=tuple(retry_http_statuses or []),
        exception_types=tuple(_get_exception_class(e) for e in (retry_exceptions or [])),
        cause_exception_types=tuple(_get_exception_class(e) for e in (retry_cause_exceptions or [])),
    )
    click.echo(f"Metering service client {retry_config=}")

    return metering.SyncMeteringServiceClient(
        base_url, token_loader=metering.auth.AuthTokenLoader(auth_config), retry_config=retry_config
    )


@object_tag("!asset-client")
def load_asset_client(
    url: str,
    credentials: Optional[ClientCredentials] = None,
    instance_name: str = "",
    request_timeout: float = 5.0,
    retries: int = 3,
) -> AssetClient:
    """Generates :class:`buildgrid_metering.client.AssetClient`
    using the tag ``!asset-client``.

    Usage
        .. code:: yaml

            - !asset-client
              url: https://remote-asset.com
              instance-name: dev
              credentials:
                tls-client-cert: /path/to/cert
                auth-token: /path/to/token
              request-timeout: 5
              retries: 3
    """

    credentials = credentials or {}
    channel, *_ = setup_channel(
        remote_url=url,
        auth_token=credentials.get("auth-token"),
        client_key=credentials.get("tls-client-key"),
        client_cert=credentials.get("tls-client-cert"),
        server_cert=credentials.get("tls-server-cert"),
        timeout=request_timeout,
    )
    return AssetClient(channel=channel, instance_name=instance_name, retries=retries)


@object_tag("!introspection")
def load_introspection_instance(scheduler: Scheduler) -> IntrospectionInstance:
    return IntrospectionInstance(scheduler)


def _parse_size(size: str) -> int:
    """Convert a string containing a size in bytes (e.g. '2GB') to a number."""
    _size_prefixes = {"k": 2**10, "m": 2**20, "g": 2**30, "t": 2**40}
    size = size.lower()

    if size[-1] == "b":
        size = size[:-1]
    if size[-1] in _size_prefixes:
        return int(size[:-1]) * _size_prefixes[size[-1]]
    return int(size)


def _validate_url_and_credentials(url: str, credentials: Optional[ClientCredentials]) -> bool:
    """Validate a URL and set of credentials for the URL.

    This parses the given URL, to determine if it should be used with
    credentials (ie. to create a secure gRPC channel), or not (ie. to create
    an insecure gRPC channel).

    ClientCredentials will be ignored for insecure channels, but if specified need
    to be valid for secure channels. Secure client channels with no specified
    credentials are valid, since gRPC will attempt to fall back to a default
    root certificate location used with no private key or certificate chain.

    If the credentials are invalid, then this function will output the error
    using ``click.echo``, and return ``False``. Otherwise this function will
    return True

    Args:
        url (str): The URL to use for validation.
        credentials (dict, optional): The credentials configuration to validate.

    """
    try:
        parsed_url = urlparse(url)
    except ValueError:
        click.echo(
            click.style(
                "ERROR: Failed to parse URL for gRPC channel construction.\n" + f"The problematic URL was: {url}.\n",
                fg="red",
                bold=True,
            ),
            err=True,
        )
        return False
    unix_socket = parsed_url.scheme == "unix"

    if parsed_url.scheme in INSECURE_URI_SCHEMES:
        # Its a URL for an insecure channel that we recognize
        if credentials is not None:
            click.echo(
                click.style(
                    "WARNING: credentials were specified for a gRPC channel, but "
                    f"`{url}` uses an insecure scheme. The credentials will be "
                    "ignored.\n",
                    fg="bright_yellow",
                )
            )
        return True

    elif parsed_url.scheme not in SECURE_URI_SCHEMES and not unix_socket:
        # Its not insecure, and its not a recognized secure scheme, so error out.
        click.echo(click.style(f"ERROR: URL {url} uses an unsupported scheme.\n", fg="red", bold=True), err=True)
        return False

    if not credentials:
        # Unix sockets are treated as secure only if credentials are set
        if not unix_socket:
            click.echo(
                click.style(
                    f"WARNING: {url} uses a secure scheme but no credentials were "
                    "specified. gRPC will attempt to fall back to defaults.\n",
                    fg="bright_yellow",
                )
            )
        return True

    client_key = credentials.get("tls-client-key")
    client_cert = credentials.get("tls-client-cert")
    server_cert = credentials.get("tls-server-cert")

    valid = True
    missing = {}
    if server_cert is not None and not os.path.exists(server_cert):
        valid = False
        missing["tls-server-cert"] = server_cert
    if client_key is not None and not os.path.exists(client_key):
        valid = False
        missing["tls-client-key"] = client_key
    if client_cert is not None and not os.path.exists(client_cert):
        valid = False
        missing["tls-client-cert"] = client_cert

    if not valid:
        click.echo(
            click.style(
                "ERROR: one or more configured TLS credentials files were "
                + "missing.\nSet remote url scheme to `http` or `grpc` in order to "
                + "deactivate TLS encryption.\nMissing files:",
                fg="red",
                bold=True,
            ),
            err=True,
        )
        for key, path in missing.items():
            click.echo(click.style(f"  - {key}: {path}", fg="red", bold=True), err=True)
        return False
    return True


def _validate_server_credentials(credentials: Optional[Dict[str, str]]) -> None:
    """Validate a configured set of credentials.

    If the credentials are invalid, then this function will call ``sys.exit``
    and stop the process, since there's no point continuing. If this function
    returns without exiting the program, then the credentials were valid.

    Args:
        credentials (dict): The credentials configuration to validate.

    """
    if not credentials:
        click.echo(
            click.style(
                "ERROR: no TLS certificates were specified for the server's network config.\n"
                + "Set `insecure-mode` to True to deactivate TLS encryption.\n",
                fg="red",
                bold=True,
            ),
            err=True,
        )
        sys.exit(-1)

    server_key = credentials.get("tls-server-key")
    server_cert = credentials.get("tls-server-cert")
    client_certs = credentials.get("tls-client-certs")

    valid = True
    missing = {}
    if server_cert is None or not os.path.exists(server_cert):
        valid = False
        missing["tls-server-cert"] = server_cert
    if server_key is None or not os.path.exists(server_key):
        valid = False
        missing["tls-server-key"] = server_key
    if client_certs is not None and not os.path.exists(client_certs):
        valid = False
        missing["tls-client-certs"] = client_certs

    if not valid:
        click.echo(
            click.style(
                "ERROR: Couldn't find certificates for secure server port.\n"
                "Set `insecure-mode` to True to deactivate TLS encryption.\n"
                "Missing files:",
                fg="red",
                bold=True,
            ),
            err=True,
        )
        for key, path in missing.items():
            click.echo(click.style(f"  - {key}: {path}", fg="red", bold=True), err=True)
        sys.exit(-1)


def get_logstream_connection_info(logstream: Any) -> Tuple[Optional[str], Optional[Dict[str, str]], Optional[str]]:
    logstream_url = None
    credentials = None
    logstream_instance_name = None
    if logstream:
        logstream_url = logstream["url"]
        credentials = logstream.get("credentials")
        if not _validate_url_and_credentials(logstream_url, credentials=credentials):
            sys.exit(-1)
        logstream_instance_name = logstream.get("instance-name", "")

    return logstream_url, credentials, logstream_instance_name


def get_schema(strict: bool = False) -> Any:
    """
    Gets a schema for the buildgrid configuration.
    If in strict mode, all object definitions will set additionalProperties to false
    """

    schema_text = files("buildgrid.server.app.settings").joinpath("schema.yml").read_text()
    schema = yaml.safe_load(schema_text)

    if strict:

        def disable_unknown_properties(item: Any) -> None:
            if isinstance(item, dict):
                if "properties" in item and "additionalProperties" not in item:
                    item["additionalProperties"] = False
                for value in item.values():
                    disable_unknown_properties(value)
            if isinstance(item, list):
                for value in item:
                    disable_unknown_properties(value)

        disable_unknown_properties(schema)

    return schema


def validate_config(path: "os.PathLike[str]", strict: bool = False, fail_deprecations: bool = False) -> None:
    """
    Validate a buildgrid configuration against its schema.
    In this mode, no real components are loaded, it simply detects invalid argument values.
    If in strict mode, all object definitions will set additionalProperties to false
    """

    with open(path) as f:
        return validate_config_value(f.read(), strict=strict, fail_deprecations=fail_deprecations)


def validate_config_value(data: str, strict: bool = False, fail_deprecations: bool = False) -> None:
    """
    Validate a buildgrid configuration against its schema.
    In this mode, no real components are loaded, it simply detects invalid argument values.
    If in strict mode, all object definitions will set additionalProperties to false
    """

    schema = get_schema(strict=strict)

    class Loader(yaml.SafeLoader):
        """
        This loader class mocks out the response value for all components by returning simple dicts.
        Additionally, each item is validated against its object definition to avoid confusing oneOf error outputs.
        """

        def struct_loader(self, node: yaml.MappingNode) -> Any:
            if node.value == "":
                args: Dict[Hashable, Any] = {"kind": node.tag}
            else:
                args = {"kind": node.tag, **self.construct_mapping(node, deep=True)}

            component_schema = {
                "$schema": schema["$schema"],
                "$ref": f"#/definitions/{node.tag[1:]}",
                "definitions": schema["definitions"],
            }
            jsonschema.validate(args, component_schema)

            if deprecation_notes := deprecated_object_keys.get(node.tag):
                for deprecated_key, deprecation_note in deprecation_notes.items():
                    if deprecated_key in args:
                        error = f"Deprecated key {deprecated_key} used in {node.tag}: {deprecation_note}"
                        click.echo(click.style(f"WARNING: {error}", fg="bright_yellow"))
                        if fail_deprecations:
                            raise ValueError(error)

            return args

        def string_loader(self, node: yaml.Node) -> Any:
            return node.value

    for kind in object_definitions:
        Loader.add_constructor(kind, Loader.struct_loader)
    for kind in string_definitions:
        Loader.add_constructor(kind, Loader.string_loader)

    instance = yaml.load(data, Loader=Loader)
    jsonschema.validate(instance, schema)


def load_config(path: "os.PathLike[str]") -> Any:
    """
    Load and validate a buildgrid configuration.
    """

    with open(path) as f:
        return load_config_value(f.read())


def load_config_value(data: str) -> Any:
    """
    Load and validate a buildgrid configuration.
    """

    validate_config_value(data, strict=False)

    class Loader(yaml.SafeLoader):
        """
        This loader class proxies tag lookups to concreate loader factory methods.
        """

        def struct_loader(self, node: yaml.MappingNode) -> Any:
            loader = object_definitions[node.tag]
            if node.value == "":
                args: Dict[str, Any] = {}
            else:
                args = {str(k): v for k, v in self.construct_mapping(node, deep=True).items()}

            for key, value in dict(args).items():
                args[key.replace("-", "_")] = args.pop(key)

            return loader(**args)

        def string_loader(self, node: yaml.Node) -> Any:
            return string_definitions[node.tag](node.value)

    for kind in object_definitions:
        Loader.add_constructor(kind, Loader.struct_loader)
    for kind in string_definitions:
        Loader.add_constructor(kind, Loader.string_loader)

    return yaml.load(data, Loader=Loader)



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add a deleted flag for Indexed CAS entries

Revision ID: 69ceb5938c06
Revises: e287a533bbd7
Create Date: 2020-03-31 13:45:29.641272

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "69ceb5938c06"
down_revision = "e287a533bbd7"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("index") as batch_op:
        batch_op.add_column(sa.Column("deleted", sa.Boolean(), nullable=False, server_default=sa.false()))


def downgrade() -> None:
    op.drop_column("index", "deleted")



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add stream resource names to jobs

Revision ID: 1553978c6e69
Revises: bbb77b202e31
Create Date: 2020-07-07 16:11:37.840129

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1553978c6e69"
down_revision = "bbb77b202e31"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("jobs", sa.Column("stdout_stream_name", sa.String(), nullable=True))
    op.add_column("jobs", sa.Column("stdout_stream_write_name", sa.String(), nullable=True))
    op.add_column("jobs", sa.Column("stderr_stream_name", sa.String(), nullable=True))
    op.add_column("jobs", sa.Column("stderr_stream_write_name", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("jobs", "stdout_stream_write_name")
    op.drop_column("jobs", "stdout_stream_name")
    op.drop_column("jobs", "stderr_stream_write_name")
    op.drop_column("jobs", "stderr_stream_name")
    # ### end Alembic commands ###
